# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Allelopathic Harvest.

This substrate contains three different varieties of berry (red, green, & blue)
and a fixed number of berry patches, which could be replanted to grow any color
variety of berry. The growth rate of each berry variety depends linearly on the
fraction that that color comprises of the total. Players have three planting
actions with which they can replant berries in their chosen color. All players
prefer to eat red berries (reward of 2 per red berry they eat versus a reward
of 1 per other colored berry). Players can achieve higher return by selecting
just one single color of berry to plant, but which one to pick is, in principle,
difficult to coordinate (start-up problem) -- though in this case all prefer
red berries, suggesting a globally rational chioce. They also always prefer to
eat berries over spending time planting (free-rider problem).

Allelopathic Harvest was first described in Koster et al. (2020).

Köster, R., McKee, K.R., Everett, R., Weidinger, L., Isaac, W.S., Hughes, E.,
Duenez-Guzman, E.A., Graepel, T., Botvinick, M. and Leibo, J.Z., 2020.
Model-free conventions in multi-agent reinforcement learning with heterogeneous
preferences. arXiv preprint arXiv:2010.09054.
"""

from typing import Any, Dict, Mapping, Sequence

from ml_collections import config_dict

from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import game_object_utils
from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

PrefabConfig = game_object_utils.PrefabConfig

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

# How many different colors of berries.
NUM_BERRY_TYPES = 3

DEFAULT_ASCII_MAP = """
333PPPP12PPP322P32PPP1P13P3P3
1PPPP2PP122PPP3P232121P2PP2P1
P1P3P11PPP13PPP31PPPP23PPPPPP
PPPPP2P2P1P2P3P33P23PP2P2PPPP
P1PPPPPPP2PPP12311PP3321PPPPP
133P2PP2PPP3PPP1PPP2213P112P1
3PPPPPPPPPPPPP31PPPPPP1P3112P
PP2P21P21P33PPPPPPP3PP2PPPP1P
PPPPP1P1P32P3PPP22PP1P2PPPP2P
PPP3PP3122211PPP2113P3PPP1332
PP12132PP1PP1P321PP1PPPPPP1P3
PPP222P12PPPP1PPPP1PPP321P11P
PPP2PPPP3P2P1PPP1P23322PP1P13
23PPP2PPPP2P3PPPP3PP3PPP3PPP2
2PPPP3P3P3PP3PP3P1P3PP11P21P1
21PPP2PP331PP3PPP2PPPPP2PP3PP
P32P2PP2P1PPPPPPP12P2PPP1PPPP
P3PP3P2P21P3PP2PP11PP1323P312
2P1PPPPP1PPP1P2PPP3P32P2P331P
PPPPP1312P3P2PPPP3P32PPPP2P11
P3PPPP221PPP2PPPPPPPP1PPP311P
32P3PPPPPPPPPP31PPPP3PPP13PPP
PPP3PPPPP3PPPPPP232P13PPPPP1P
P1PP1PPP2PP3PPPPP33321PP2P3PP
P13PPPP1P333PPPP2PP213PP2P3PP
1PPPPP3PP2P1PP21P3PPPP231P2PP
1331P2P12P2PPPP2PPP3P23P21PPP
P3P131P3PPP13P1PPP222PPPP11PP
2P3PPPPPPPP2P323PPP2PPP1PPP2P
21PPPPPPP12P23P1PPPPPP13P3P11
"""
SPRITE_SIZE = 8

# Map a character to the prefab it represents in the ASCII map.
CHAR_PREFAB_MAP = {
    "P": {"type": "all", "list": ["floor", "spawn_point"]},
    "W": "wall",
    "1": {"type": "all", "list": ["soil", "berry_1"]},
    "2": {"type": "all", "list": ["soil", "berry_2"]},
    "3": {"type": "all", "list": ["soil", "berry_3"]},
}


# These need to be orthogonal, same intensity and variance.
COLORS = [
    (200, 10, 10, 255),  # 'Red'
    (10, 200, 10, 255),  # 'Green'
    (10, 10, 200, 255),  # 'Blue'
]

ROLE_TO_MOST_TASTY_BERRY_IDX = {
    "player_who_likes_red": 0,
    "player_who_likes_green": 1,
    "player_who_likes_blue": 2,
}

MARKING_SPRITE = """
oxxxxxxo
xoxxxxox
xxoxxoxx
xxxooxxx
xxxooxxx
xxoxxoxx
xoxxxxox
oxxxxxxo
"""


def get_marking_palette(alpha: float) -> Dict[str, Sequence[int]]:
  alpha_uint8 = int(alpha * 255)
  assert alpha_uint8 >= 0.0 and alpha_uint8 <= 255, "Color value out of range."
  return {"x": shapes.ALPHA, "o": (0, 0, 0, alpha_uint8)}

_NUM_DIRECTIONS = 4  # NESW

FLOOR = {
    "name": "floor",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "floor",
                "stateConfigs": [{
                    "state": "floor",
                    "layer": "background",
                    "sprite": "Floor",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Floor",],
                "spriteShapes": [shapes.DIRT_PATTERN],
                "palettes": [{
                    "x": (55, 55, 55, 255),
                    "X": (60, 60, 60, 255),
                }],
                "noRotates": [True]
            }
        },
        {
            "component": "Transform",
        },
    ]
}

SOIL = {
    "name": "soil",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "soil",
                "stateConfigs": [{
                    "state": "soil",
                    "layer": "background",
                    "sprite": "Soil",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Soil",],
                "spriteShapes": [shapes.SOIL],
                "palettes": [{
                    "D": (40, 40, 40, 255),
                    "d": (50, 50, 50, 255),
                    "X": (60, 60, 60, 255),
                    "x": (70, 70, 70, 255)}],
                "noRotates": [False]
            }
        },
        {
            "component": "Transform",
        },
    ]
}

WALL = {
    "name": "wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall",
                "stateConfigs": [{
                    "state": "wall",
                    "layer": "upperPhysical",
                    "sprite": "Wall",
                }],
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "spriteNames": ["Wall"],
                # This color is a dark shade of grey.
                "spriteRGBColors": [(40, 40, 40)]
            }
        },
        {
            "component": "BeamBlocker",
            "kwargs": {
                "beamType": "zapHit"
            }
        },
        {
            "component": "BeamBlocker",
            "kwargs": {
                "beamType": "fire_1"
            }
        },
        {
            "component": "BeamBlocker",
            "kwargs": {
                "beamType": "fire_2"
            }
        },
        {
            "component": "BeamBlocker",
            "kwargs": {
                "beamType": "fire_3"
            }
        },
    ]
}

SPAWN_POINT = {
    "name": "spawnPoint",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "spawnPoint",
                "stateConfigs": [{
                    "state": "spawnPoint",
                    "layer": "logic",
                    "groups": ["spawnPoints"]
                }],
            }
        },
        {
            "component": "Transform",
        },
    ]
}


def create_berry_prefab(lua_index: int):
  """Return a berry prefab for the given color index (initial color)."""
  berry = {
      "name": "berry",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "unripe_{}".format(lua_index),
                  "stateConfigs": [
                      # Unripe states.
                      {
                          "state": "unripe_1",
                          "layer": "lowerPhysical",
                          "sprite": "UnripeBerry_1",
                          "groups": ["unripes"]
                      },
                      {
                          "state": "unripe_2",
                          "layer": "lowerPhysical",
                          "sprite": "UnripeBerry_2",
                          "groups": ["unripes"]
                      },
                      {
                          "state": "unripe_3",
                          "layer": "lowerPhysical",
                          "sprite": "UnripeBerry_3",
                          "groups": ["unripes"]
                      },
                      # Ripe states.
                      {
                          "state": "ripe_1",
                          "layer": "lowerPhysical",
                          "sprite": "RipeBerry_1",
                          "groups": []
                      },
                      {
                          "state": "ripe_2",
                          "layer": "lowerPhysical",
                          "sprite": "RipeBerry_2",
                          "groups": []
                      },
                      {
                          "state": "ripe_3",
                          "layer": "lowerPhysical",
                          "sprite": "RipeBerry_3",
                          "groups": []
                      },
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [
                      "UnripeBerry_1",
                      "UnripeBerry_2",
                      "UnripeBerry_3",
                      "RipeBerry_1",
                      "RipeBerry_2",
                      "RipeBerry_3",
                  ],
                  "spriteShapes": [
                      shapes.BERRY_SEEDS,
                      shapes.BERRY_SEEDS,
                      shapes.BERRY_SEEDS,
                      shapes.BERRY_RIPE,
                      shapes.BERRY_RIPE,
                      shapes.BERRY_RIPE,
                  ],
                  "palettes": [
                      # Unripe colors
                      {
                          "o": COLORS[0],
                          "O": shapes.scale_color(COLORS[0], 1.5),
                          "x": (0, 0, 0, 0)
                      },
                      {
                          "o": COLORS[1],
                          "O": shapes.scale_color(COLORS[1], 1.5),
                          "x": (0, 0, 0, 0)
                      },
                      {
                          "o": COLORS[2],
                          "O": shapes.scale_color(COLORS[2], 1.5),
                          "x": (0, 0, 0, 0)
                      },
                      # Ripe colors
                      {
                          "d": COLORS[0],
                          "O": shapes.scale_color(COLORS[0], 1.5),
                          "o": shapes.scale_color(COLORS[0], 1.25),
                          "x": (0, 0, 0, 0)
                      },
                      {
                          "d": COLORS[1],
                          "O": shapes.scale_color(COLORS[1], 1.5),
                          "o": shapes.scale_color(COLORS[1], 1.25),
                          "x": (0, 0, 0, 0)
                      },
                      {
                          "d": COLORS[2],
                          "O": shapes.scale_color(COLORS[2], 1.5),
                          "o": shapes.scale_color(COLORS[2], 1.25),
                          "x": (0, 0, 0, 0)
                      },
                  ],
                  # Note: the berries do not rotate in this version (unlike in
                  # the original allelopathic_harvest version, where they do).
                  "noRotates": [True] * (NUM_BERRY_TYPES * 2)
              }
          },
          {
              "component": "Berry",
              "kwargs": {
                  "unripePrefix": "unripe",
                  "ripePrefix": "ripe",
                  "colorId": lua_index,
              }
          },
          {
              "component": "Edible",
              "kwargs": {
                  "name": "Edible",
                  "eatingSetsColorToNewborn": True,
              }
          },
          {
              "component": "Regrowth",
              "kwargs": {
                  "minimumTimeToRipen": 10,
                  "baseRate": 5e-6,
                  "linearGrowth": True,
              }
          },
          {
              "component": "Coloring",
              "kwargs": {
                  "numColors": NUM_BERRY_TYPES,
              }
          },
      ]
  }
  return berry


def create_avatar_object(player_idx: int,
                         most_tasty_berry_idx: int) -> Dict[str, Any]:
  """Return the avatar for the player numbered `player_idx`."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1

  lua_most_tasty_berry_idx = most_tasty_berry_idx + 1

  live_state_name = "player{}".format(lua_index)
  avatar_sprite_name = "avatarSprite{}".format(lua_index)
  avatar_object = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": live_state_name,
                  "stateConfigs": [
                      # Initial player state.
                      {"state": live_state_name,
                       "layer": "upperPhysical",
                       "sprite": avatar_sprite_name,
                       "contact": "avatar",
                       "groups": ["players"]},

                      # Player wait type for when they have been zapped out.
                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [avatar_sprite_name],
                  "spriteShapes": [shapes.CUTE_AVATAR],
                  # This color is white. It should never appear in gameplay. So
                  # if a white colored avatar does appear then something is
                  # broken.
                  "palettes": [shapes.get_palette((255, 255, 255))],
                  "noRotates": [True]
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "aliveState": live_state_name,
                  "waitState": "playerWait",
                  "speed": 1.0,
                  "spawnGroup": "spawnPoints",
                  "actionOrder": ["move",
                                  "turn",
                                  "fireZap",
                                  "fire_1",
                                  "fire_2",
                                  "fire_3"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": _NUM_DIRECTIONS},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "fireZap": {"default": 0, "min": 0, "max": 1},
                      "fire_1": {"default": 0, "min": 0, "max": 1},
                      "fire_2": {"default": 0, "min": 0, "max": 1},
                      "fire_3": {"default": 0, "min": 0, "max": 1},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  },
              }
          },
          {
              "component": "Zapper",
              "kwargs": {
                  "cooldownTime": 4,
                  "beamLength": 3,
                  "beamRadius": 1,
                  "beamColor": (253, 253, 253),  # the zapper beam is white.
                  "framesTillRespawn": 25,
                  "penaltyForBeingZapped": 0,  # leave this always at 0.
                  "rewardForZapping": 0,  # leave this always at 0.
                  # GraduatedSanctionsMarking handles removal instead of Zapper.
                  "removeHitPlayer": False,
              }
          },
          {
              "component": "ReadyToShootObservation",
          },
          {
              "component": "Taste",
              "kwargs": {
                  "mostTastyBerryId": lua_most_tasty_berry_idx,
                  "rewardMostTasty": 2,
              }
          },
          {
              "component": "ColorZapper",
              "kwargs": {
                  "cooldownTime": 2,
                  "beamLength": 3,
                  "beamRadius": 0,
                  "numColorZappers": NUM_BERRY_TYPES,
                  "beamColors": COLORS,
                  # When `eatingSetsColorToNewborn` and `stochasticallyCryptic`
                  # are both true than stochastically change back to the
                  # newborn color after eating a berry with probability
                  # inversely related to the monoculture fraction. So larger
                  # monoculture fractions yield lower probabilities of changing
                  # back to the newborn color.
                  "stochasticallyCryptic": True,
              }
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })
    avatar_object["components"].append({
        "component": "AvatarMetricReporter",
        "kwargs": {
            "metrics": [
                {
                    "name": "COLOR_ID",
                    "type": "Doubles",
                    "shape": [],
                    "component": "ColorZapper",
                    "variable": "colorId",
                },
                {
                    "name": "MOST_TASTY_BERRY_ID",
                    "type": "Doubles",
                    "shape": [],
                    "component": "Taste",
                    "variable": "mostTastyBerryId",
                },
            ]
        },
    })
    avatar_object["components"].append({
        "component": "AvatarIdsInViewObservation",
    })
    avatar_object["components"].append({
        "component": "AvatarIdsInRangeToZapObservation",
    })
  return avatar_object

# PREFABS is a dictionary mapping names to template game objects that can
# be cloned and placed in multiple locations accoring to an ascii map.
PREFABS = {
    "floor": FLOOR,
    "soil": SOIL,
    "wall": WALL,
    "spawn_point": SPAWN_POINT,
    "berry_1": create_berry_prefab(1),
    "berry_2": create_berry_prefab(2),
    "berry_3": create_berry_prefab(3),
}

# PLAYER_COLOR_PALETTES is a list with each entry specifying the color to use
# for the player at the corresponding index.
NUM_PLAYERS_UPPER_BOUND = 60
PLAYER_COLOR_PALETTES = []
for i in range(NUM_PLAYERS_UPPER_BOUND):
  PLAYER_COLOR_PALETTES.append(shapes.get_palette(colors.palette[i]))

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "fireZap": 0, "fire_1": 0, "fire_2": 0, "fire_3": 0}
FORWARD    = {"move": 1, "turn":  0, "fireZap": 0, "fire_1": 0, "fire_2": 0, "fire_3": 0}
STEP_RIGHT = {"move": 2, "turn":  0, "fireZap": 0, "fire_1": 0, "fire_2": 0, "fire_3": 0}
BACKWARD   = {"move": 3, "turn":  0, "fireZap": 0, "fire_1": 0, "fire_2": 0, "fire_3": 0}
STEP_LEFT  = {"move": 4, "turn":  0, "fireZap": 0, "fire_1": 0, "fire_2": 0, "fire_3": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "fireZap": 0, "fire_1": 0, "fire_2": 0, "fire_3": 0}
TURN_RIGHT = {"move": 0, "turn":  1, "fireZap": 0, "fire_1": 0, "fire_2": 0, "fire_3": 0}
FIRE_ZAP   = {"move": 0, "turn":  0, "fireZap": 1, "fire_1": 0, "fire_2": 0, "fire_3": 0}
FIRE_ONE   = {"move": 0, "turn":  0, "fireZap": 0, "fire_1": 1, "fire_2": 0, "fire_3": 0}
FIRE_TWO   = {"move": 0, "turn":  0, "fireZap": 0, "fire_1": 0, "fire_2": 1, "fire_3": 0}
FIRE_THREE = {"move": 0, "turn":  0, "fireZap": 0, "fire_1": 0, "fire_2": 0, "fire_3": 1}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    FIRE_ZAP,
    FIRE_ONE,
    FIRE_TWO,
    FIRE_THREE,
)


# The Scene objece is a non-physical object, it components implement global
# logic. In this case, that includes holding the global berry counters to
# implement the regrowth rate, as well as some of the observations.
def create_scene(num_players: int):
  """Creates the global scene."""
  scene = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "GlobalBerryTracker",
              "kwargs": {
                  "numBerryTypes": NUM_BERRY_TYPES,
                  "numPlayers": num_players,
              }
          },
          {
              "component": "GlobalZapTracker",
              "kwargs": {
                  "numBerryTypes": NUM_BERRY_TYPES,
                  "numPlayers": num_players,
              }
          },
          {
              "component": "GlobalMetricHolder",
              "kwargs": {
                  "metrics": [
                      {"type": "tensor.Int32Tensor",
                       "shape": (num_players, num_players),
                       "variable": "playerZapMatrix"},
                  ]
              }
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    scene["components"].append({
        "component": "GlobalMetricReporter",
        "kwargs": {
            "metrics": [
                {
                    "name": "RIPE_BERRIES_BY_TYPE",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES,),
                    "component": "GlobalBerryTracker",
                    "variable": "ripeBerriesPerType",
                },
                {
                    "name": "UNRIPE_BERRIES_BY_TYPE",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES,),
                    "component": "GlobalBerryTracker",
                    "variable": "unripeBerriesPerType",
                },
                {
                    "name": "BERRIES_BY_TYPE",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES,),
                    "component": "GlobalBerryTracker",
                    "variable": "berriesPerType",
                },
                {
                    "name": "COLORING_BY_PLAYER",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES, num_players),
                    "component": "GlobalBerryTracker",
                    "variable": "coloringByPlayerMatrix",
                },
                {
                    "name": "EATING_TYPES_BY_PLAYER",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES, num_players),
                    "component": "GlobalBerryTracker",
                    "variable": "eatingTypesByPlayerMatrix",
                },
                {
                    "name": "BERRIES_PER_TYPE_BY_COLOR_OF_COLORER",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES, NUM_BERRY_TYPES + 1),
                    "component": "GlobalBerryTracker",
                    "variable": "berryTypesByColorOfColorer",
                },
                {
                    "name": "BERRIES_PER_TYPE_BY_TASTE_OF_COLORER",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES, NUM_BERRY_TYPES),
                    "component": "GlobalBerryTracker",
                    "variable": "berryTypesByTasteOfColorer",
                },
                {
                    "name": "PLAYER_TIMEOUT_COUNT",
                    "type": "tensor.Int32Tensor",
                    "shape": (num_players, num_players),
                    "component": "GlobalZapTracker",
                    "variable": "fullZapCountMatrix",
                },
                {
                    "name": "COLOR_BY_COLOR_ZAP_COUNTS",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES + 1, NUM_BERRY_TYPES + 1),
                    "component": "GlobalZapTracker",
                    "variable": "colorByColorZapCounts",
                },
                {
                    "name": "COLOR_BY_TASTE_ZAP_COUNTS",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES + 1, NUM_BERRY_TYPES),
                    "component": "GlobalZapTracker",
                    "variable": "colorByTasteZapCounts",
                },
                {
                    "name": "TASTE_BY_TASTE_ZAP_COUNTS",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES, NUM_BERRY_TYPES),
                    "component": "GlobalZapTracker",
                    "variable": "tasteByTasteZapCounts",
                },
                {
                    "name": "TASTE_BY_COLOR_ZAP_COUNTS",
                    "type": "tensor.Int32Tensor",
                    "shape": (NUM_BERRY_TYPES, NUM_BERRY_TYPES + 1),
                    "component": "GlobalZapTracker",
                    "variable": "tasteByColorZapCounts",
                },
                {
                    "name": "WHO_ZAPPED_WHO",
                    "type": "tensor.Int32Tensor",
                    "shape": (num_players, num_players),
                    "component": "GlobalMetricHolder",
                    "variable": "playerZapMatrix",
                },
            ]
        }
    })
  return scene


def create_marking_overlay(player_idx: int) -> Dict[str, Any]:
  """Create a graduated sanctions marking overlay object."""
  # Lua is 1-indexed.
  lua_idx = player_idx + 1

  marking_object = {
      "name": "avatar_marking",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "avatarMarkingWait",
                  "stateConfigs": [
                      # Declare one state per level of the hit logic.
                      {"state": "level_1",
                       "layer": "superOverlay",
                       "sprite": "sprite_for_level_1"},
                      {"state": "level_2",
                       "layer": "superOverlay",
                       "sprite": "sprite_for_level_2"},
                      {"state": "level_3",
                       "layer": "superOverlay",
                       "sprite": "sprite_for_level_3"},

                      # Invisible inactive (zapped out) overlay type.
                      {"state": "avatarMarkingWait",
                       "groups": ["avatarMarkingWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["sprite_for_level_1",
                                  "sprite_for_level_2",
                                  "sprite_for_level_3"],
                  "spriteShapes": [MARKING_SPRITE,
                                   MARKING_SPRITE,
                                   MARKING_SPRITE],
                  "palettes": [get_marking_palette(0.0),
                               get_marking_palette(0.5),
                               get_marking_palette(1.0)],
                  "noRotates": [True] * 3
              }
          },
          {
              "component": "GraduatedSanctionsMarking",
              "kwargs": {
                  "playerIndex": lua_idx,
                  "waitState": "avatarMarkingWait",
                  "hitName": "zapHit",
                  "recoveryTime": 50,
                  "hitLogic": [
                      {"levelIncrement": 1,
                       "sourceReward": 0,
                       "targetReward": 0,
                       "freeze": 25},
                      {"levelIncrement": -1,
                       "sourceReward": 0,
                       "targetReward": -10,
                       "remove": True}
                  ],
              }
          },
      ]
  }
  return marking_object


def create_colored_avatar_overlay(player_idx: int) -> Dict[str, Any]:
  """Create a colored avatar overlay object."""
  # Lua is 1-indexed.
  lua_idx = player_idx + 1

  overlay_object = {
      "name": "avatar_overlay",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "avatarOverlayWait",
                  "stateConfigs": [
                      # Invisible active overlay type.
                      {
                          "state": "avatarOverlay",
                          "layer": "overlay",
                          "sprite": "NewbornAvatar",
                          "groups": ["avatarOverlays"]
                      },

                      # Invisible inactive (zapped out) overlay type.
                      {
                          "state": "avatarOverlayWait",
                          "groups": ["avatarOverlayWaits"]
                      },

                      # Colored overlay piece types for use after the player has
                      # colored a berry with a coloring beam.
                      {
                          "state": "coloredPlayer_1",
                          "layer": "overlay",
                          "sprite": "ColoredAvatar_1",
                          "groups": ["avatarOverlays"]
                      },
                      {
                          "state": "coloredPlayer_2",
                          "layer": "overlay",
                          "sprite": "ColoredAvatar_2",
                          "groups": ["avatarOverlays"]
                      },
                      {
                          "state": "coloredPlayer_3",
                          "layer": "overlay",
                          "sprite": "ColoredAvatar_3",
                          "groups": ["avatarOverlays"]
                      },
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["NewbornAvatar"] + [
                      "ColoredAvatar_{}".format(i)
                      for i in range(1, NUM_BERRY_TYPES + 1)
                  ],
                  "spriteShapes": [shapes.CUTE_AVATAR] *
                                  (NUM_BERRY_TYPES + 1),
                  "palettes":
                      [shapes.get_palette((125, 125, 125))] +
                      [shapes.get_palette(beam_color) for beam_color in COLORS],
                  "noRotates": [True] * (NUM_BERRY_TYPES + 1)
              }
          },
          {
              "component": "AvatarConnector",
              "kwargs": {
                  "playerIndex": lua_idx,
                  "aliveState": "avatarOverlay",
                  "waitState": "avatarOverlayWait"
              }
          },
      ]
  }
  return overlay_object


def create_avatar_and_associated_objects(
    roles: Sequence[str]):
  """Returns list of avatar objects and associated other objects."""
  avatar_objects = []
  additional_objects = []
  for player_idx, role in enumerate(roles):
    if role == "default":
      most_tasty_berry_idx = player_idx % 2
    else:
      most_tasty_berry_idx = ROLE_TO_MOST_TASTY_BERRY_IDX[role]

    avatar_object = create_avatar_object(
        player_idx=player_idx, most_tasty_berry_idx=most_tasty_berry_idx)
    avatar_objects.append(avatar_object)

    overlay_object = create_colored_avatar_overlay(player_idx)
    marking_object = create_marking_overlay(player_idx)
    additional_objects.append(overlay_object)
    additional_objects.append(marking_object)

  return avatar_objects + additional_objects


def get_config():
  """Default configuration for the allelopathic harvest level."""
  config = config_dict.ConfigDict()

  config.episode_timesteps = 2000
  config.ascii_map = DEFAULT_ASCII_MAP

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "READY_TO_SHOOT",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(ACTION_SET))
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.world_rgb(DEFAULT_ASCII_MAP, SPRITE_SIZE),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"default",
                                  "player_who_likes_red",
                                  "player_who_likes_green",
                                  "player_who_likes_blue",})

  return config


def build(
    roles: Sequence[str],
    config: config_dict.ConfigDict,
) -> Mapping[str, Any]:
  """Build the allelopathic_harvest substrate given roles."""
  num_players = len(roles)
  game_objects = create_avatar_and_associated_objects(roles=roles)
  # Build the rest of the substrate definition.
  substrate_definition = dict(
      levelName="allelopathic_harvest",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      maxEpisodeLengthFrames=config.episode_timesteps,
      spriteSize=SPRITE_SIZE,
      topology="TORUS",  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": config.ascii_map,
          "gameObjects": game_objects,
          "scene": create_scene(num_players),
          "prefabs": PREFABS,
          "charPrefabMap": CHAR_PREFAB_MAP,
          "playerPalettes": [PLAYER_COLOR_PALETTES[0]] * num_players,
      },
  )
  return substrate_definition



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for discrete_action_wrapper."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import dm_env
import dmlab2d
import numpy as np

from meltingpot.utils.substrates.wrappers import discrete_action_wrapper

MOVE_SPEC = dm_env.specs.BoundedArray(
    shape=(), minimum=0, maximum=3, dtype=np.int8, name='MOVE')
TURN_SPEC = dm_env.specs.BoundedArray(
    shape=(), minimum=0, maximum=3, dtype=np.int8, name='TURN')
VALID_VALUE_0 = np.zeros([], dtype=np.int8)
VALID_VALUE_1 = np.array(3, dtype=np.int8)
INVALID_VALUE = np.array(4, dtype=np.int8)


class Lab2DToListsWrapperTest(parameterized.TestCase):

  def test_valid_set(self):
    env = mock.Mock(spec_set=dmlab2d.Environment)
    env.action_spec.return_value = [
        {'MOVE': MOVE_SPEC, 'TURN': TURN_SPEC},
        {'MOVE': MOVE_SPEC, 'TURN': TURN_SPEC},
    ]
    discrete_action_wrapper.Wrapper(env, action_table=[
        {'MOVE': VALID_VALUE_0, 'TURN': VALID_VALUE_0},
        {'MOVE': VALID_VALUE_0, 'TURN': VALID_VALUE_1},
        {'MOVE': VALID_VALUE_1, 'TURN': VALID_VALUE_0},
        {'MOVE': VALID_VALUE_1, 'TURN': VALID_VALUE_1},
    ])

  @parameterized.named_parameters(
      ('empty', []),
      ('out_of_bounds', [{'MOVE': INVALID_VALUE, 'TURN': VALID_VALUE_0}]),
      ('missing_key', [{'TURN': VALID_VALUE_0}]),
      ('extra_key', [{'INVALID': VALID_VALUE_0}]),
  )
  def test_invalid_set(self, action_table):
    env = mock.Mock(spec_set=dmlab2d.Environment)
    env.action_spec.return_value = [
        {'MOVE': MOVE_SPEC, 'TURN': TURN_SPEC},
        {'MOVE': MOVE_SPEC, 'TURN': TURN_SPEC},
    ]
    with self.assertRaises(ValueError):
      discrete_action_wrapper.Wrapper(env, action_table=action_table)

  def test_action_spec(self):
    env = mock.Mock(spec_set=dmlab2d.Environment)
    env.action_spec.return_value = [
        {'MOVE': MOVE_SPEC, 'TURN': TURN_SPEC},
        {'MOVE': MOVE_SPEC, 'TURN': TURN_SPEC},
    ]
    wrapped = discrete_action_wrapper.Wrapper(env, action_table=[
        {'MOVE': VALID_VALUE_0, 'TURN': VALID_VALUE_0},
        {'MOVE': VALID_VALUE_0, 'TURN': VALID_VALUE_1},
        {'MOVE': VALID_VALUE_1, 'TURN': VALID_VALUE_0},
    ])
    actual = wrapped.action_spec()
    expected = (
        dm_env.specs.DiscreteArray(num_values=3, dtype=np.int64, name='action'),
    ) * 2
    self.assertEqual(actual, expected)

  def test_step(self):
    env = mock.Mock(spec_set=dmlab2d.Environment)
    env.action_spec.return_value = [
        {'MOVE': MOVE_SPEC, 'TURN': TURN_SPEC},
        {'MOVE': MOVE_SPEC, 'TURN': TURN_SPEC},
    ]
    env.step.return_value = mock.sentinel.timestep
    wrapped = discrete_action_wrapper.Wrapper(env, action_table=[
        {'MOVE': VALID_VALUE_0, 'TURN': VALID_VALUE_0},
        {'MOVE': VALID_VALUE_0, 'TURN': VALID_VALUE_1},
        {'MOVE': VALID_VALUE_1, 'TURN': VALID_VALUE_0},
    ])
    actual = wrapped.step([0, 2])

    with self.subTest('timestep'):
      np.testing.assert_equal(actual, mock.sentinel.timestep)
    with self.subTest('action'):
      (action,), _ = env.step.call_args
      self.assertEqual(action, [
          {'MOVE': VALID_VALUE_0, 'TURN': VALID_VALUE_0},
          {'MOVE': VALID_VALUE_1, 'TURN': VALID_VALUE_0},
      ])


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Install script for setuptools."""

import os
import shutil
import tarfile
import urllib.request

import setuptools
from setuptools.command import build_py

VERSION = '2.1.1-dev.11'
ASSETS_VERSION = '2.1.0'

ASSETS_URL = f'http://storage.googleapis.com/dm-meltingpot/meltingpot-assets-{ASSETS_VERSION}.tar.gz'


class BuildPy(build_py.build_py):
  """Command that downloads Melting Pot assets as part of build_py."""

  def run(self):
    self.download_and_extract_assets()
    if not self.editable_mode:
      super().run()
      self.build_assets()

  def download_and_extract_assets(self):
    """Downloads and extracts assets to meltingpot/assets."""
    tar_file_path = os.path.join(
        self.get_package_dir('assets'), os.path.basename(ASSETS_URL))
    if os.path.exists(tar_file_path):
      print(f'found cached assets {tar_file_path}', flush=True)
    else:
      os.makedirs(os.path.dirname(tar_file_path), exist_ok=True)
      print('downloading assets...', flush=True)
      urllib.request.urlretrieve(ASSETS_URL, filename=tar_file_path)
      print(f'downloaded {tar_file_path}', flush=True)

    root = os.path.join(self.get_package_dir(''), 'meltingpot')
    os.makedirs(root, exist_ok=True)
    if os.path.exists(f'{root}/assets'):
      shutil.rmtree(f'{root}/assets')
      print('deleted existing assets', flush=True)
    with tarfile.open(tar_file_path, mode='r|*') as tarball:
      tarball.extractall(root)
    print(f'extracted assets from {tar_file_path} to {root}/assets', flush=True)

  def build_assets(self):
    """Copies assets from package to build lib."""
    package_root = os.path.join(self.get_package_dir(''), 'meltingpot')
    os.makedirs(package_root, exist_ok=True)
    build_root = os.path.join(self.build_lib, 'meltingpot')
    if os.path.exists(f'{build_root}/assets'):
      shutil.rmtree(f'{build_root}/assets')
      print('deleted existing assets', flush=True)
    shutil.copytree(f'{package_root}/assets', f'{build_root}/assets')
    print(f'copied assets from {package_root}/assets to {build_root}/assets',
          flush=True)


setuptools.setup(
    name='dm-meltingpot',
    version=VERSION,
    license='Apache 2.0',
    license_files=['LICENSE'],
    url='https://github.com/deepmind/meltingpot',
    download_url='https://github.com/deepmind/meltingpot/releases',
    author='DeepMind',
    author_email='noreply@google.com',
    description=(
        'A suite of test scenarios for multi-agent reinforcement learning.'),
    keywords='multi-agent reinforcement-learning python machine-learning',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    cmdclass={'build_py': BuildPy},
    package_dir={
        'meltingpot': 'meltingpot',
    },
    package_data={
        'meltingpot.lua': ['**'],
    },
    python_requires='>=3.10',
    install_requires=[
        'absl-py',
        'chex<0.1.81',  # Incompatible with tensorflow 2.13 (due to numpy req).
        'dm-env',
        'dmlab2d',
        'dm-tree',
        'immutabledict',
        'ml-collections',
        'networkx',
        'numpy',
        'opencv-python',
        'pandas',
        'pygame',
        'reactivex',
        'tensorflow',
    ],
    extras_require={
        # Used in development.
        'dev': [
            'build',
            'isort',
            'pipreqs',
            'pyink',
            'pylint',
            'pytest-xdist',
            'pytype',
        ],
    },
)



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Commons Harvest: Open.

Example video: https://youtu.be/lZ-qpPP4BNE

Apples are spread around the map and can be consumed for a reward of 1. Apples
that have been consumed regrow with a per-step probability that depends on the
number of uneaten apples in a `L2` norm neighborhood of radius 2 (by default).
After an apple has been eaten and thus removed, its regrowth probability depends
on the number of uneaten apples still in its local neighborhood. With standard
parameters, it the grown rate decreases as the number of uneaten apples in the
neighborhood decreases and when there are zero uneaten apples in the
neighborhood then the regrowth rate is zero. As a consequence, a patch of apples
that collectively doesn't have any nearby apples, can be irrevocably lost if all
apples in the patch are consumed. Therefore, agents must exercise restraint when
consuming apples within a patch. Notice that in a single agent situation, there
is no incentive to collect the last apple in a patch (except near the end of the
episode). However, in a multi-agent situation, there is an incentive for any
agent to consume the last apple rather than risk another agent consuming it.
This creates a tragedy of the commons from which the substrate derives its name.

This mechanism was first described in Janssen et al (2010) and adapted for
multi-agent reinforcement learning in Perolat et al (2017).

Janssen, M.A., Holahan, R., Lee, A. and Ostrom, E., 2010. Lab experiments for
the study of social-ecological systems. Science, 328(5978), pp.613-617.

Perolat, J., Leibo, J.Z., Zambaldi, V., Beattie, C., Tuyls, K. and Graepel, T.,
2017. A multi-agent reinforcement learning model of common-pool
resource appropriation. In Proceedings of the 31st International Conference on
Neural Information Processing Systems (pp. 3646-3655).
"""

from typing import Any, Dict, Mapping, Sequence

from ml_collections import config_dict
import numpy as np

from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

APPLE_RESPAWN_RADIUS = 2.0
REGROWTH_PROBABILITIES = [0.0, 0.0025, 0.005, 0.025]

ASCII_MAP = """
WWWWWWWWWWWWWWWWWWWWWWWW
WAAA    A      A    AAAW
WAA    AAA    AAA    AAW
WA    AAAAA  AAAAA    AW
W      AAA    AAA      W
W       A      A       W
W  A                A  W
W AAA  Q        Q  AAA W
WAAAAA            AAAAAW
W AAA              AAA W
W  A                A  W
W                      W
W                      W
W                      W
W  PPPPPPPPPPPPPPPPPP  W
W PPPPPPPPPPPPPPPPPPPP W
WPPPPPPPPPPPPPPPPPPPPPPW
WWWWWWWWWWWWWWWWWWWWWWWW
"""

# `prefab` determines which prefab game object to use for each `char` in the
# ascii map.
CHAR_PREFAB_MAP = {
    "P": {"type": "all", "list": ["floor", "spawn_point"]},
    "Q": {"type": "all", "list": ["floor", "inside_spawn_point"]},
    " ": "floor",
    "W": "wall",
    "A": {"type": "all", "list": ["grass", "apple"]},
}

_COMPASS = ["N", "E", "S", "W"]

FLOOR = {
    "name": "floor",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "floor",
                "stateConfigs": [{
                    "state": "floor",
                    "layer": "background",
                    "sprite": "Floor",
                }],
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Floor"],
                "spriteShapes": [shapes.GRAINY_FLOOR],
                "palettes": [{"*": (220, 205, 185, 255),
                              "+": (210, 195, 175, 255),}],
                "noRotates": [False]
            }
        },
    ]
}

GRASS = {
    "name":
        "grass",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "grass",
                "stateConfigs": [
                    {
                        "state": "grass",
                        "layer": "background",
                        "sprite": "Grass"
                    },
                    {
                        "state": "dessicated",
                        "layer": "background",
                        "sprite": "Floor"
                    },
                ],
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Grass", "Floor"],
                "spriteShapes": [
                    shapes.GRASS_STRAIGHT, shapes.GRAINY_FLOOR
                ],
                "palettes": [{
                    "*": (158, 194, 101, 255),
                    "@": (170, 207, 112, 255)
                }, {
                    "*": (220, 205, 185, 255),
                    "+": (210, 195, 175, 255),
                }],
                "noRotates": [False, False]
            }
        },
    ]
}

WALL = {
    "name": "wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall",
                "stateConfigs": [{
                    "state": "wall",
                    "layer": "upperPhysical",
                    "sprite": "Wall",
                }],
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Wall"],
                "spriteShapes": [shapes.WALL],
                "palettes": [{"*": (95, 95, 95, 255),
                              "&": (100, 100, 100, 255),
                              "@": (109, 109, 109, 255),
                              "#": (152, 152, 152, 255)}],
                "noRotates": [False]
            }
        },
        {
            "component": "BeamBlocker",
            "kwargs": {
                "beamType": "zapHit"
            }
        },
    ]
}

SPAWN_POINT = {
    "name": "spawnPoint",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "spawnPoint",
                "stateConfigs": [{
                    "state": "spawnPoint",
                    "layer": "alternateLogic",
                    "groups": ["spawnPoints"]
                }],
            }
        },
        {
            "component": "Transform",
        },
    ]
}

INSIDE_SPAWN_POINT = {
    "name": "spawnPoint",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "spawnPoint",
                "stateConfigs": [{
                    "state": "spawnPoint",
                    "layer": "alternateLogic",
                    "groups": ["insideSpawnPoints"]
                }],
            }
        },
        {
            "component": "Transform",
        },
    ]
}

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "fireZap": 0}
FORWARD    = {"move": 1, "turn":  0, "fireZap": 0}
STEP_RIGHT = {"move": 2, "turn":  0, "fireZap": 0}
BACKWARD   = {"move": 3, "turn":  0, "fireZap": 0}
STEP_LEFT  = {"move": 4, "turn":  0, "fireZap": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "fireZap": 0}
TURN_RIGHT = {"move": 0, "turn":  1, "fireZap": 0}
FIRE_ZAP   = {"move": 0, "turn":  0, "fireZap": 1}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    FIRE_ZAP,
)

TARGET_SPRITE_SELF = {
    "name": "Self",
    "shape": shapes.CUTE_AVATAR,
    "palette": shapes.get_palette((50, 100, 200)),
    "noRotate": True,
}


def create_scene():
  """Creates the scene with the provided args controlling apple regrowth."""
  scene = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Neighborhoods",
              "kwargs": {}
          },
          {
              "component": "StochasticIntervalEpisodeEnding",
              "kwargs": {
                  "minimumFramesPerEpisode": 1000,
                  "intervalLength": 100,  # Set equal to unroll length.
                  "probabilityTerminationPerInterval": 0.15
              }
          }
      ]
  }

  return scene


def create_apple_prefab(regrowth_radius=-1.0,  # pylint: disable=dangerous-default-value
                        regrowth_probabilities=[0, 0.0, 0.0, 0.0]):
  """Creates the apple prefab with the provided settings."""
  growth_rate_states = [
      {
          "state": "apple",
          "layer": "lowerPhysical",
          "sprite": "Apple",
          "groups": ["apples"]
      },
      {
          "state": "appleWait",
          "layer": "logic",
          "sprite": "AppleWait",
      },
  ]
  # Enumerate all possible states for a potential apple. There is one state for
  # each regrowth rate i.e., number of nearby apples.
  upper_bound_possible_neighbors = np.floor(np.pi*regrowth_radius**2+1)+1
  for i in range(int(upper_bound_possible_neighbors)):
    growth_rate_states.append(dict(state="appleWait_{}".format(i),
                                   layer="logic",
                                   groups=["waits_{}".format(i)],
                                   sprite="AppleWait"))

  apple_prefab = {
      "name": "apple",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "apple",
                  "stateConfigs": growth_rate_states,
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["Apple", "AppleWait"],
                  "spriteShapes": [shapes.APPLE, shapes.FILL],
                  "palettes": [
                      {"x": (0, 0, 0, 0),
                       "*": (214, 88, 88, 255),
                       "#": (194, 79, 79, 255),
                       "o": (53, 132, 49, 255),
                       "|": (102, 51, 61, 255)},
                      {"i": (0, 0, 0, 0)}],
                  "noRotates": [True, True]
              }
          },
          {
              "component": "Edible",
              "kwargs": {
                  "liveState": "apple",
                  "waitState": "appleWait",
                  "rewardForEating": 1.0,
              }
          },
          {
              "component": "DensityRegrow",
              "kwargs": {
                  "liveState": "apple",
                  "waitState": "appleWait",
                  "radius": regrowth_radius,
                  "regrowthProbabilities": regrowth_probabilities,
              }
          },
      ]
  }

  return apple_prefab


def create_prefabs(regrowth_radius=-1.0,
                   # pylint: disable=dangerous-default-value
                   regrowth_probabilities=[0, 0.0, 0.0, 0.0]):
  """Returns a dictionary mapping names to template game objects."""
  prefabs = {
      "floor": FLOOR,
      "grass": GRASS,
      "wall": WALL,
      "spawn_point": SPAWN_POINT,
      "inside_spawn_point": INSIDE_SPAWN_POINT,
  }
  prefabs["apple"] = create_apple_prefab(
      regrowth_radius=regrowth_radius,
      regrowth_probabilities=regrowth_probabilities)
  return prefabs


def create_avatar_object(player_idx: int,
                         target_sprite_self: Dict[str, Any],
                         spawn_group: str) -> Dict[str, Any]:
  """Create an avatar object that always sees itself as blue."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1

  # Setup the self vs other sprite mapping.
  source_sprite_self = "Avatar" + str(lua_index)
  custom_sprite_map = {source_sprite_self: target_sprite_self["name"]}

  live_state_name = "player{}".format(lua_index)
  avatar_object = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": live_state_name,
                  "stateConfigs": [
                      {"state": live_state_name,
                       "layer": "upperPhysical",
                       "sprite": source_sprite_self,
                       "contact": "avatar",
                       "groups": ["players"]},

                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [source_sprite_self],
                  "spriteShapes": [shapes.CUTE_AVATAR],
                  "palettes": [shapes.get_palette(
                      colors.human_readable[player_idx])],
                  "noRotates": [True]
              }
          },
          {
              "component": "AdditionalSprites",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "customSpriteNames": [target_sprite_self["name"]],
                  "customSpriteShapes": [target_sprite_self["shape"]],
                  "customPalettes": [target_sprite_self["palette"]],
                  "customNoRotates": [target_sprite_self["noRotate"]],
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "aliveState": live_state_name,
                  "waitState": "playerWait",
                  "speed": 1.0,
                  "spawnGroup": spawn_group,
                  "postInitialSpawnGroup": "spawnPoints",
                  "actionOrder": ["move", "turn", "fireZap"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": len(_COMPASS)},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "fireZap": {"default": 0, "min": 0, "max": 1},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  },
                  "spriteMap": custom_sprite_map,
              }
          },
          {
              "component": "Zapper",
              "kwargs": {
                  "cooldownTime": 2,
                  "beamLength": 3,
                  "beamRadius": 1,
                  "framesTillRespawn": 4,
                  "penaltyForBeingZapped": 0,
                  "rewardForZapping": 0,
              }
          },
          {
              "component": "ReadyToShootObservation",
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })

  return avatar_object


def create_avatar_objects(num_players):
  """Returns list of avatar objects of length 'num_players'."""
  avatar_objects = []
  for player_idx in range(0, num_players):
    spawn_group = "spawnPoints"
    if player_idx < 2:
      # The first two player slots always spawn closer to the apples.
      spawn_group = "insideSpawnPoints"

    game_object = create_avatar_object(player_idx,
                                       TARGET_SPRITE_SELF,
                                       spawn_group=spawn_group)
    avatar_objects.append(game_object)

  return avatar_objects


def get_config():
  """Default configuration for training on the commons_harvest level."""
  config = config_dict.ConfigDict()

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "READY_TO_SHOOT",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(ACTION_SET))
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(144, 192),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"default"})
  config.default_player_roles = ("default",) * 7

  return config


def build(
    roles: Sequence[str],
    config: config_dict.ConfigDict,
) -> Mapping[str, Any]:
  """Build substrate definition given player roles."""
  del config
  num_players = len(roles)
  # Build the rest of the substrate definition.
  substrate_definition = dict(
      levelName="commons_harvest",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      # Define upper bound of episode length since episodes end stochastically.
      maxEpisodeLengthFrames=5000,
      spriteSize=8,
      topology="BOUNDED",  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": ASCII_MAP,
          "gameObjects": create_avatar_objects(num_players),
          "prefabs": create_prefabs(APPLE_RESPAWN_RADIUS,
                                    REGROWTH_PROBABILITIES),
          "charPrefabMap": CHAR_PREFAB_MAP,
          "scene": create_scene(),
      },
  )
  return substrate_definition



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for coins puppeteers."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized

from meltingpot.utils.puppeteers import coins
from meltingpot.utils.puppeteers import testutils

_COOPERATE = mock.sentinel.cooperate
_DEFECT = mock.sentinel.defect
_SPITE = mock.sentinel.spite
_NUM_DEFECTIONS_KEY = 'DEFECTIONS'


def _goals(puppeteer, num_defections, state=None):
  observations = [{_NUM_DEFECTIONS_KEY: n} for n in num_defections]
  goals, state = testutils.goals_from_observations(puppeteer, observations,
                                                   state)
  return goals, state


class ReciprocatorTest(parameterized.TestCase):

  def test_threshold(self):
    puppeteer = coins.Reciprocator(
        cooperate_goal=_COOPERATE,
        defect_goal=_DEFECT,
        spite_goal=_SPITE,
        partner_defection_signal=_NUM_DEFECTIONS_KEY,
        recency_window=1,
        threshold=3,
        frames_to_punish=1,
        spiteful_punishment_window=0,
    )
    num_defections = [0, 1, 2, 3]
    expected = [_COOPERATE, _COOPERATE, _COOPERATE, _DEFECT]
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)

  @parameterized.parameters(
      [(1, 0, 0, 1), (_COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE)],
      [(1, 0, 1), (_COOPERATE, _COOPERATE, _DEFECT)],
      [(1, 1), (_COOPERATE, _DEFECT)],
  )
  def test_recency_window(self, num_defections, expected):
    puppeteer = coins.Reciprocator(
        cooperate_goal=_COOPERATE,
        defect_goal=_DEFECT,
        spite_goal=_SPITE,
        partner_defection_signal=_NUM_DEFECTIONS_KEY,
        recency_window=3,
        threshold=2,
        frames_to_punish=1,
        spiteful_punishment_window=0,
    )
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)

  @parameterized.parameters(1, 2, 3)
  def test_defect_duration(self, duration):
    puppeteer = coins.Reciprocator(
        cooperate_goal=_COOPERATE,
        defect_goal=_DEFECT,
        spite_goal=_SPITE,
        partner_defection_signal=_NUM_DEFECTIONS_KEY,
        recency_window=1,
        threshold=1,
        frames_to_punish=duration,
        spiteful_punishment_window=1,
    )
    num_defections = [1, 0, 0, 0]
    expected = (
        [_SPITE] + [_DEFECT] * (duration - 1) + [_COOPERATE] * (4 - duration))
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)

  @parameterized.parameters(1, 2, 4)
  def test_spite_duration(self, duration):
    puppeteer = coins.Reciprocator(
        cooperate_goal=_COOPERATE,
        defect_goal=_DEFECT,
        spite_goal=_SPITE,
        partner_defection_signal=_NUM_DEFECTIONS_KEY,
        recency_window=1,
        threshold=1,
        frames_to_punish=4,
        spiteful_punishment_window=duration,
    )
    num_defections = [1, 0, 0, 0]
    expected = [_SPITE] * duration + [_DEFECT] * (4 - duration)
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)

  def test_resets_on_first(self):
    puppeteer = coins.Reciprocator(
        cooperate_goal=_COOPERATE,
        defect_goal=_DEFECT,
        spite_goal=_SPITE,
        partner_defection_signal=_NUM_DEFECTIONS_KEY,
        recency_window=8,
        threshold=1,
        frames_to_punish=8,
        spiteful_punishment_window=8,
    )
    _, state = _goals(puppeteer, [0, 0, 1, 0])
    num_defections = [0, 0, 0, 0]
    expected = [_COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE]
    actual, _ = _goals(puppeteer, num_defections, state)
    self.assertSequenceEqual(actual, expected)

  def test_defection_during_defect_resets_spite(self):
    puppeteer = coins.Reciprocator(
        cooperate_goal=_COOPERATE,
        defect_goal=_DEFECT,
        spite_goal=_SPITE,
        partner_defection_signal=_NUM_DEFECTIONS_KEY,
        recency_window=1,
        threshold=1,
        frames_to_punish=3,
        spiteful_punishment_window=1,
    )
    num_defections = [1, 0, 1, 0, 0, 0]
    expected = [_SPITE, _DEFECT, _SPITE, _DEFECT, _DEFECT, _COOPERATE]
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)

  def test_defection_during_spite_extends_spite(self):
    puppeteer = coins.Reciprocator(
        cooperate_goal=_COOPERATE,
        defect_goal=_DEFECT,
        spite_goal=_SPITE,
        partner_defection_signal=_NUM_DEFECTIONS_KEY,
        recency_window=1,
        threshold=1,
        frames_to_punish=3,
        spiteful_punishment_window=2,
    )
    num_defections = [1, 0, 1, 0, 0, 0]
    expected = [_SPITE, _SPITE, _SPITE, _SPITE, _DEFECT, _COOPERATE]
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)

  def test_impulse_response(self):
    puppeteer = coins.Reciprocator(
        cooperate_goal=_COOPERATE,
        defect_goal=_DEFECT,
        spite_goal=_SPITE,
        partner_defection_signal=_NUM_DEFECTIONS_KEY,
        recency_window=4,
        threshold=1,
        frames_to_punish=2,
        spiteful_punishment_window=1,
    )
    num_defections = [1, 0, 0, 0]
    expected = [_SPITE, _DEFECT, _COOPERATE, _COOPERATE]
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)

  def test_boxcar_response(self):
    puppeteer = coins.Reciprocator(
        cooperate_goal=_COOPERATE,
        defect_goal=_DEFECT,
        spite_goal=_SPITE,
        partner_defection_signal=_NUM_DEFECTIONS_KEY,
        recency_window=4,
        threshold=1,
        frames_to_punish=2,
        spiteful_punishment_window=1,
    )
    num_defections = [1, 1, 1, 0, 0, 0]
    expected = [_SPITE, _SPITE, _SPITE, _DEFECT, _COOPERATE, _COOPERATE]
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for observables_wrapper."""

import dataclasses
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import dmlab2d

from meltingpot.utils.substrates.wrappers import observables_wrapper


class ObservablesWrapperTest(parameterized.TestCase):

  def test_observables(self):
    base = mock.create_autospec(
        dmlab2d.Environment, instance=True, spec_set=True)
    with observables_wrapper.ObservablesWrapper(base) as env:
      received = []
      observables = env.observables()
      for field in dataclasses.fields(observables):
        getattr(observables, field.name).subscribe(
            on_next=received.append,
            on_error=lambda e: received.append(type(e)),
            on_completed=lambda: received.append('DONE'),
        )

      base.reset.return_value = mock.sentinel.timestep_0
      base.events.return_value = [mock.sentinel.events_0]
      env.reset()
      base.step.return_value = mock.sentinel.timestep_1
      base.events.return_value = [mock.sentinel.events_1]
      env.step(mock.sentinel.action_1)
      base.step.return_value = mock.sentinel.timestep_2
      base.events.return_value = [mock.sentinel.events_2]
      env.step(mock.sentinel.action_2)

    self.assertSequenceEqual(received, [
        mock.sentinel.timestep_0,
        mock.sentinel.events_0,
        mock.sentinel.action_1,
        mock.sentinel.timestep_1,
        mock.sentinel.events_1,
        mock.sentinel.action_2,
        mock.sentinel.timestep_2,
        mock.sentinel.events_2,
        'DONE',
        'DONE',
        'DONE',
    ])


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for the substrate: fruit_market_concentric_rivers.

Example video: https://youtu.be/djmylRv1i_w

This substrate has three concentric rings of water that confer a small stamina
cost to players who step on them.
"""

from ml_collections import config_dict as configdict

from meltingpot.configs.substrates import fruit_market as base_config
from meltingpot.utils.substrates import specs

build = base_config.build

ASCII_MAP = """
xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
x/___________________________+x
x'###########################`x
x!~~~~~~~~~~~~~~~~~~~~~~~~~~~|x
x!~~~~~~~~~~~~~~~~~~~~~~~~~~~|x
x!~~~LLLLLLLLLLLLLLLLLLLLL~~~|x
x!~~~L~~~~~~~~~~~~~~~~~~~L~~~|x
x!~~~L~~~~~~~~~~~~~~~~~~~L~~~|x
x!~~~L~~LLLLLLLLLLLLLLL~~L~~~|x
x!~~~L~~L~~~~~~~~~~~~~L~~L~~~|x
x!~~~L~~L~~~~~~~~~~~~~L~~L~~~|x
x!~~~L~~L~~LLLLLLLLL~~L~~L~~~|x
x!~~~L~~L~~LP~P~P~PL~~L~~L~~~|x
x!~~~L~~L~~L~P~P~P~L~~L~~L~~~|x
x!~~~L~~L~~L~~P~P~~L~~L~~L~~~|x
x!~~~L~~L~~L~P~P~P~L~~L~~L~~~|x
x!~~~L~~L~~L~~P~P~~L~~L~~L~~~|x
x!~~~L~~L~~L~P~P~P~L~~L~~L~~~|x
x!~~~L~~L~~LP~P~P~PL~~L~~L~~~|x
x!~~~L~~L~~LLLLLLLLL~~L~~L~~~|x
x!~~~L~~L~~~~~~~~~~~~~L~~L~~~|x
x!~~~L~~L~~~~~~~~~~~~~L~~L~~~|x
x!~~~L~~LLLLLLLLLLLLLLL~~L~~~|x
x!~~~L~~~~~~~~~~~~~~~~~~~L~~~|x
x!~~~L~~~~~~~~~~~~~~~~~~~L~~~|x
x!~~~LLLLLLLLLLLLLLLLLLLLL~~~|x
x!~~~~~~~~~~~~~~~~~~~~~~~~~~~|x
x!~~~~~~~~~~~~~~~~~~~~~~~~~~~|x
x!~~~~~~~~~~~~~~~~~~~~~~~~~~~|x
x(---------------------------)x
x<###########################>x
"""

# Map a character to the prefab it represents in the ASCII map.
CHAR_PREFAB_MAP = {
    # wall prefabs
    "/": {"type": "all", "list": ["ground", "nw_wall_corner"]},
    "'": {"type": "all", "list": ["ground", "nw_inner_wall_corner"]},
    "+": {"type": "all", "list": ["ground", "ne_wall_corner"]},
    "`": {"type": "all", "list": ["ground", "ne_inner_wall_corner"]},
    ")": {"type": "all", "list": ["ground", "se_wall_corner"]},
    "(": {"type": "all", "list": ["ground", "sw_wall_corner"]},
    "_": {"type": "all", "list": ["ground", "wall_north"]},
    "|": {"type": "all", "list": ["ground", "wall_east"]},
    "-": {"type": "all", "list": ["ground", "wall_south"]},
    "!": {"type": "all", "list": ["ground", "wall_west"]},
    "#": {"type": "all", "list": ["ground", "wall_shadow_s"]},
    ">": {"type": "all", "list": ["ground", "wall_shadow_se"]},
    "<": {"type": "all", "list": ["ground", "wall_shadow_sw"]},

    # non-wall prefabs
    "L": "river",
    "P": {"type": "all", "list": ["ground", "potential_tree", "spawn_point"]},
    "~": {"type": "all", "list": ["ground", "potential_tree"]},
    "x": "ground",
}


def get_config():
  """Configuration for this substrate."""
  config = base_config.get_config()
  # Specify the number of players to particate in each episode (optional).
  config.recommended_num_players = 16
  # Override the map layout settings.
  config.layout = configdict.ConfigDict()
  config.layout.ascii_map = ASCII_MAP
  config.layout.char_prefab_map = CHAR_PREFAB_MAP

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(base_config.ACTION_SET))
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      "STAMINA": specs.float64(),
      "INVENTORY": specs.int64(2),
      "MY_OFFER": specs.int64(2),
      "OFFERS": specs.int64(102),
      "HUNGER": specs.float64(),
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(248, 248,),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"apple_farmer", "banana_farmer"})
  config.default_player_roles = ("apple_farmer",) * 8 + ("banana_farmer",) * 8
  return config



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Puppeteers for puppet bots."""

import abc
from typing import Generic, Mapping, NewType, Sequence, Tuple, TypeVar

import dm_env
import immutabledict
import numpy as np

State = TypeVar('State')
PuppetGoal = NewType('PuppetGoal', np.ndarray)

_GOAL_OBSERVATION_KEY = 'GOAL'
_GOAL_DTYPE = np.int32


class Puppeteer(Generic[State], metaclass=abc.ABCMeta):
  """A puppeteer that controls the timestep forwarded to the puppet.

  Must not possess any mutable state not in `initial_state`.
  """

  @abc.abstractmethod
  def initial_state(self) -> State:
    """Returns the initial state of the puppeteer.

    Must not have any side effects.
    """

  @abc.abstractmethod
  def step(self, timestep: dm_env.TimeStep,
           prev_state: State) -> Tuple[dm_env.TimeStep, State]:
    """Steps the puppeteer.

    Must not have any side effects.

    Args:
      timestep: information from the environment.
      prev_state: the previous state of the puppeteer.

    Returns:
      timestep: the timestep to forward to the puppet.
      next_state: the state for the next step call.
    """


def puppet_timestep(timestep: dm_env.TimeStep,
                    goal: PuppetGoal) -> dm_env.TimeStep:
  """Returns a timestep with a goal observation added."""
  puppet_observation = immutabledict.immutabledict(
      timestep.observation, **{_GOAL_OBSERVATION_KEY: goal})
  return timestep._replace(observation=puppet_observation)


def puppet_goals(names: Sequence[str],
                 dtype: ... = _GOAL_DTYPE) -> Mapping[str, PuppetGoal]:
  """Returns a mapping from goal name to a one-hot goal vector for a puppet.

  Args:
    names: names for each of the corresponding goals.
    dtype: dtype of the one-hot goals to return.
  """
  goals = np.eye(len(names), dtype=dtype)
  goals.setflags(write=False)
  return immutabledict.immutabledict(zip(names, goals))



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Capture the Flag.

Example video: https://youtu.be/ECzevYpi1dM

This substrate a team based zero sum game. There are four players on each team.

There is a red team and blue team. Players  can paint the ground anywhere by
using their zapping beam. If they stand on their own color then they gain health
up to a maximum of 3 (so they are more likely to win shootouts). They lose
health down to 1 from their default of 2 when standing on the opposing team's
color (so they are more likely to lose shootouts in that case). Health recovers
stochastically, at a fixed rate of 0.05 per frame. It cannot exceed its maximum,
determined by the current color of the ground the agent is standing on.

Players also cannot move over their opposing team's color. If the opposing team
paints the square underneath their feet then they get stuck in place until they
use their own zapping beam to re-paint the square underneath and in front of
themselves to break free. In practice this slows them down by one frame (which
may be critical if they are being chased).

Friendly fire is impossible; agents cannot zap their teammates.

In the _Capture the Flag_ substrate the final goal is capturing the opposing
team's flag. Payoffs are common to the entire winning team. Indicator tiles
around the edge of the map and in its very center display which teams have their
own flag on their base, allowing them the possibility of capturing their
opponent's flag by bringing it to their own base/flag. When indicator tiles are
red then only the red team can score. When indicator tiles are blue then only
the blue team can score. When the indicator tiles are purple then both teams
have the possibility of scoring (though neither is close to doing so) since both
flags are in their respective home bases.
"""

from typing import Any, Dict, Mapping, Optional, Sequence

from ml_collections import config_dict
import numpy as np

from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

_COMPASS = ["N", "E", "S", "W"]

ASCII_MAP = """
IIIIIIIIIIIIIIIIIIIIIII
IWWWWWWWWWWWWWWWWWWWWWI
IWPPP,PPPP,F,PPPP,PPPWI
IWPPP,,PP,,,,,PP,,PPPWI
IWPPP,,,,,,,,,,,,,PPPWI
IWP,,WW,,,,,,,,,WW,,PWI
IWHHWWW,WWWWWWW,WWWHHWI
IWHHW,D,,,,,,,,,D,WHHWI
IWHH,,W,,,WWW,,,W,,HHWI
IW,,,,W,,,,,,,,,W,,,,WI
IW,,,,WWW,,,,,WWW,,,,WI
IW,,,,,,,,,I,,,,,,,,,WI
IW,,,,WWW,,,,,WWW,,,,WI
IW,,,,W,,,,,,,,,W,,,,WI
IWHH,,W,,,WWW,,,W,,HHWI
IWHHW,D,,,,,,,,,D,WHHWI
IWHHWWW,WWWWWWW,WWWHHWI
IWQ,,WW,,,,,,,,,WW,,QWI
IWQQQ,,,,,,,,,,,,,QQQWI
IWQQQ,,QQ,,,,,QQ,,QQQWI
IWQQQ,QQQQ,G,QQQQ,QQQWI
IWWWWWWWWWWWWWWWWWWWWWI
IIIIIIIIIIIIIIIIIIIIIII
"""

# `prefab` determines which prefab game object to use for each `char` in the
# ascii map.
CHAR_PREFAB_MAP = {
    "P": {"type": "all", "list": ["spawn_point_red", "ground"]},
    "Q": {"type": "all", "list": ["spawn_point_blue", "ground"]},
    "W": "wall",
    "D": {"type": "choice",
          "list": ["destroyable_wall"] * 9 + ["destroyed_wall"]},
    "H": {"type": "choice",
          "list": ["destroyable_wall"] * 3 + ["destroyed_wall"]},
    ",": "ground",
    "I": {"type": "all", "list": ["indicator", "indicator_frame"]},
    "F": {"type": "all", "list": ["ground", "home_tile_red", "flag_red"]},
    "G": {"type": "all", "list": ["ground", "home_tile_blue", "flag_blue"]},
}

RED_COLOR = (225, 55, 85, 255)
DARKER_RED_COLOR = (200, 35, 55, 255)
DARKEST_RED_COLOR = (160, 5, 25, 255)

BLUE_COLOR = (85, 55, 225, 255)
DARKER_BLUE_COLOR = (55, 35, 200, 255)
DARKEST_BLUE_COLOR = (25, 5, 160, 255)

PURPLE_COLOR = (107, 63, 160, 255)


def multiply_tuple(color_tuple, factor):
  alpha = color_tuple[3]
  return tuple([int(np.min([x * factor, alpha])) for x in color_tuple[0: 3]])

TEAMS_DATA = {
    "red": {"color": RED_COLOR,
            "spawn_group": "{}SpawnPoints".format("red")},
    "blue": {"color": BLUE_COLOR,
             "spawn_group": "{}SpawnPoints".format("blue")},
}

WALL = {
    "name": "wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall",
                "stateConfigs": [{
                    "state": "wall",
                    "layer": "upperPhysical",
                    "sprite": "Wall",
                }],
            }
        },
        {"component": "Transform",},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Wall",],
                "spriteShapes": [shapes.WALL],
                "palettes": [{"*": (95, 95, 95, 255),
                              "&": (100, 100, 100, 255),
                              "@": (109, 109, 109, 255),
                              "#": (152, 152, 152, 255)}],
                "noRotates": [True]
            }
        },
        {
            "component": "AllBeamBlocker",
            "kwargs": {}
        },
    ]
}

INDICATOR_FRAME = {
    "name": "indicator_frame",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "inert",
                "stateConfigs": [
                    {"state": "inert",
                     "layer": "superOverlay",
                     "sprite": "InertFrame"}
                ]
            }
        },
        {"component": "Transform",},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["InertFrame"],
                "spriteShapes": [shapes.BUTTON],
                "palettes": [{"*": (0, 0, 0, 0),
                              "x": (55, 55, 55, 255),
                              "#": (0, 0, 0, 0)}],
                "noRotates": [True]
            }
        },
    ]
}


INDICATOR = {
    "name": "control_indicator",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "both",
                "stateConfigs": [
                    {
                        "state": "neither",
                        "layer": "background",
                        "sprite": "NeitherIndicator",
                    },
                    {
                        "state": "red",
                        "layer": "background",
                        "sprite": "RedIndicator",
                    },
                    {
                        "state": "blue",
                        "layer": "background",
                        "sprite": "BlueIndicator",
                    },
                    {
                        "state": "both",
                        "layer": "background",
                        "sprite": "BothIndicator",
                    },
                ]
            }
        },
        {"component": "Transform",},
        {
            "component": "Appearance",
            "kwargs": {
                "spriteNames": ["NeitherIndicator",
                                "RedIndicator",
                                "BlueIndicator",
                                "BothIndicator"],
                "spriteRGBColors": [(0, 0, 0, 0),
                                    DARKER_RED_COLOR,
                                    DARKER_BLUE_COLOR,
                                    PURPLE_COLOR]
            }
        },
        {"component": "ControlIndicator",},
    ]
}


def create_home_tile_prefab(team: str):
  """Return a home tile prefab, where the flag starts and must be brought."""
  sprite_name = "HomeTileFrame{}".format(team)
  prefab = {
      "name": "home_tile",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "inert",
                  "stateConfigs": [
                      {"state": "inert",
                       "layer": "background",
                       "sprite": sprite_name}
                  ]
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [sprite_name],
                  "spriteShapes": [shapes.BUTTON],
                  "palettes": [{"*": (0, 0, 0, 0),
                                "x": (0, 0, 0, 0),
                                "#": (218, 165, 32, 255)}],
                  "noRotates": [True]
              }
          },
          {
              "component": "HomeTile",
              "kwargs": {
                  "team": team,
              }
          },
      ]
  }
  return prefab


def create_ground_prefab():
  """Return a prefab for a colorable ground prefab."""
  sprite_names = ["RedGround", "BlueGround"]
  sprite_colors = [DARKEST_RED_COLOR, DARKEST_BLUE_COLOR]
  prefab = {
      "name": "ground",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "clean",
                  "stateConfigs": [
                      {
                          "state": "clean",
                          "layer": "alternateLogic",
                      },
                      {
                          "state": "red",
                          "layer": "alternateLogic",
                          "sprite": sprite_names[0],
                      },
                      {
                          "state": "blue",
                          "layer": "alternateLogic",
                          "sprite": sprite_names[1],
                      },
                  ]
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "spriteNames": sprite_names,
                  "spriteRGBColors": sprite_colors
              }
          },
          {
              "component": "Ground",
              "kwargs": {
                  "teamNames": ["red", "blue"],
              }
          },
      ]
  }
  return prefab


def create_destroyable_wall_prefab(initial_state):
  """Return destroyable wall prefab, potentially starting in destroyed state."""
  if initial_state == "destroyed":
    initial_health = 0
  else:
    initial_health = 5
  prefab = {
      "name": "destroyableWall",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": initial_state,
                  "stateConfigs": [
                      {
                          "state": "destroyable",
                          "layer": "upperPhysical",
                          "sprite": "DestroyableWall",
                      },
                      {
                          "state": "damaged",
                          "layer": "upperPhysical",
                          "sprite": "DamagedWall",
                      },
                      {
                          "state": "destroyed",
                          "layer": "alternateLogic",
                          "sprite": "Rubble",
                      },
                  ],
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["DestroyableWall",
                                  "DamagedWall",
                                  "Rubble"],
                  "spriteShapes": [shapes.WALL,
                                   shapes.WALL,
                                   shapes.WALL],
                  "palettes": [{"*": (55, 55, 55, 255),
                                "&": (100, 100, 100, 255),
                                "@": (109, 109, 109, 255),
                                "#": (152, 152, 152, 255)},
                               {"*": (55, 55, 55, 255),
                                "&": (100, 100, 100, 255),
                                "@": (79, 79, 79, 255),
                                "#": (152, 152, 152, 255)},
                               {"*": (0, 0, 0, 255),
                                "&": (0, 0, 0, 255),
                                "@": (29, 29, 29, 255),
                                "#": (0, 0, 0, 255)}],
                  "noRotates": [True] * 3
              }
          },
          {
              "component": "Destroyable",
              "kwargs": {"hitNames": ["red", "blue"],
                         "initialHealth": initial_health,
                         "damagedHealthLevel": 2}
          }
      ]
  }
  return prefab


def create_spawn_point_prefab(team):
  """Return a team-specific spawn-point prefab."""
  prefab = {
      "name": "spawn_point",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "playerSpawnPoint",
                  "stateConfigs": [{
                      "state": "playerSpawnPoint",
                      "layer": "logic",
                      "groups": [TEAMS_DATA[team]["spawn_group"]],
                  }],
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "invisible",
                  "spriteNames": [],
                  "spriteRGBColors": []
              }
          },
      ]
  }
  return prefab


def create_flag_prefab(team: str):
  """Return a team-specific flag prefab."""
  dropped_sprite_name = "DroppedFlag_{}".format(team)
  carried_sprite_name = "CarriedFlag_{}".format(team)
  if team == "red":
    flag_color = RED_COLOR
  elif team == "blue":
    flag_color = BLUE_COLOR

  prefab = {
      "name": "{}_flag".format(team),
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "dropped",
                  "stateConfigs": [
                      {
                          "state": "dropped",
                          "layer": "lowerPhysical",
                          "sprite": dropped_sprite_name,
                      },
                      {
                          "state": "carried",
                          "layer": "overlay",
                          "sprite": carried_sprite_name,
                      },
                      {
                          "state": "wait",
                      }
                  ]
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [dropped_sprite_name, carried_sprite_name],
                  "spriteShapes": [shapes.FLAG,
                                   shapes.FLAG_HELD],
                  "palettes": [shapes.get_palette(flag_color)] * 2,
                  "noRotates": [True, True]
              }
          },
          {
              "component": "Flag",
              "kwargs": {
                  "team": team,
              }
          }
      ]
  }
  return prefab

# PREFABS is a dictionary mapping names to template game objects that can
# be cloned and placed in multiple locations accoring to an ascii map.
PREFABS = {
    "wall": WALL,
    "spawn_point_red": create_spawn_point_prefab("red"),
    "spawn_point_blue": create_spawn_point_prefab("blue"),
    "destroyable_wall": create_destroyable_wall_prefab("destroyable"),
    "destroyed_wall": create_destroyable_wall_prefab("destroyed"),
    "ground": create_ground_prefab(),
    "indicator": INDICATOR,
    "indicator_frame": INDICATOR_FRAME,
    "flag_red": create_flag_prefab("red"),
    "flag_blue": create_flag_prefab("blue"),
    "home_tile_red": create_home_tile_prefab("red"),
    "home_tile_blue": create_home_tile_prefab("blue"),
}

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "fireZap": 0}
FORWARD    = {"move": 1, "turn":  0, "fireZap": 0}
STEP_RIGHT = {"move": 2, "turn":  0, "fireZap": 0}
BACKWARD   = {"move": 3, "turn":  0, "fireZap": 0}
STEP_LEFT  = {"move": 4, "turn":  0, "fireZap": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "fireZap": 0}
TURN_RIGHT = {"move": 0, "turn":  1, "fireZap": 0}
FIRE_ZAP_A = {"move": 0, "turn":  0, "fireZap": 1}
FIRE_ZAP_B = {"move": 0, "turn":  0, "fireZap": 2}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    FIRE_ZAP_A,  # a short-range beam with a wide area of effect
    FIRE_ZAP_B,  # a longer range beam with a thin area of effect
)


# The Scene is a non-physical object, its components implement global logic.
def create_scene():
  """Creates the global scene."""
  scene = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {"component": "Transform",},
          {
              "component": "FlagManager",
              "kwargs": {}
          },
      ]
  }
  return scene


def create_avatar_object(
    player_idx: int,
    team: str,
    override_taste_kwargs: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
  """Create an avatar object."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1

  team_color = TEAMS_DATA[team]["color"]

  health1_avatar_sprite_name = "avatarSprite{}Health1".format(lua_index)
  health2_avatar_sprite_name = "avatarSprite{}Health2".format(lua_index)
  health3_avatar_sprite_name = "avatarSprite{}Health3".format(lua_index)

  health1_color_palette = shapes.get_palette(multiply_tuple(team_color, 0.35))
  health2_color_palette = shapes.get_palette(team_color)
  health3_color_palette = shapes.get_palette(multiply_tuple(team_color, 1.75))

  taste_kwargs = {
      "defaultTeamReward": 1.0,
      "rewardForZapping": 0.0,
      "extraRewardForZappingFlagCarrier": 0.0,
      "rewardForReturningFlag": 0.0,
      "rewardForPickingUpOpposingFlag": 0.0,
  }
  if override_taste_kwargs:
    taste_kwargs.update(override_taste_kwargs)

  avatar_object = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "health2",
                  "stateConfigs": [
                      {"state": "health1",
                       "layer": "upperPhysical",
                       "sprite": health1_avatar_sprite_name,
                       "contact": "avatar",
                       "groups": ["players"]},
                      {"state": "health2",
                       "layer": "upperPhysical",
                       "sprite": health2_avatar_sprite_name,
                       "contact": "avatar",
                       "groups": ["players"]},
                      {"state": "health3",
                       "layer": "upperPhysical",
                       "sprite": health3_avatar_sprite_name,
                       "contact": "avatar",
                       "groups": ["players"]},

                      # Player wait state used when they have been zapped out.
                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [health1_avatar_sprite_name,
                                  health2_avatar_sprite_name,
                                  health3_avatar_sprite_name],
                  "spriteShapes": [shapes.CUTE_AVATAR,
                                   shapes.CUTE_AVATAR,
                                   shapes.CUTE_AVATAR],
                  "palettes": [health1_color_palette,
                               health2_color_palette,
                               health3_color_palette],
                  "noRotates": [True] * 3
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "aliveState": "health2",
                  "additionalLiveStates": ["health1", "health3"],
                  "waitState": "playerWait",
                  "spawnGroup": TEAMS_DATA[team]["spawn_group"],
                  "actionOrder": ["move",
                                  "turn",
                                  "fireZap"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": len(_COMPASS)},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "fireZap": {"default": 0, "min": 0, "max": 2},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  },
                  # The following kwarg makes it possible to get rewarded for
                  # team rewards even when an avatar is "dead".
                  "skipWaitStateRewards": False,
              }
          },
          {
              "component": "ColorZapper",
              "kwargs": {
                  "team": team,
                  # The color zapper beam is somewhat transparent.
                  "color": (team_color[0], team_color[1], team_color[2], 150),
                  "cooldownTime": 2,
                  "beamLength": 3,
                  "beamRadius": 1,
                  "secondaryBeamCooldownTime": 4,
                  "secondaryBeamLength": 6,
                  "secondaryBeamRadius": 0,
                  "aliveStates": ["health1", "health2", "health3"],
              }
          },
          {
              "component": "ReadyToShootObservation",
              "kwargs": {
                  "zapperComponent": "ColorZapper",
              }
          },
          {
              "component": "ZappedByColor",
              "kwargs": {
                  "team": team,
                  "allTeamNames": ["red", "blue"],
                  "framesTillRespawn": 80,
                  "penaltyForBeingZapped": 0,
                  "rewardForZapping": 0,
                  "healthRegenerationRate": 0.05,
                  "maxHealthOnGround": 2,
                  "maxHealthOnOwnColor": 3,
                  "maxHealthOnEnemyColor": 1,
                  "groundLayer": "alternateLogic",
              }
          },
          {
              "component": "TeamMember",
              "kwargs": {"team": team}
          },
          {
              "component": "Taste",
              "kwargs": taste_kwargs
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })

  return avatar_object


def _even_vs_odd_team_assignment(num_players,
                                 taste_kwargs: Optional[Any] = None):
  """Assign players with even ids to red team and odd ids to blue team."""
  avatar_objects = []
  for player_idx in range(0, num_players):
    if player_idx % 2 == 0:
      team = "red"
    elif player_idx % 2 == 1:
      team = "blue"
    game_object = create_avatar_object(player_idx, team,
                                       override_taste_kwargs=taste_kwargs)
    avatar_objects.append(game_object)

  return avatar_objects


def _low_vs_high_team_assignment(num_players,
                                 taste_kwargs: Optional[Any] = None):
  """Assign players with id below the median id to blue and above it to red."""
  median = np.median(range(num_players))
  avatar_objects = []
  for player_idx in range(0, num_players):
    if player_idx < median:
      team = "blue"
    elif player_idx > median:
      team = "red"
    game_object = create_avatar_object(player_idx, team,
                                       override_taste_kwargs=taste_kwargs)
    avatar_objects.append(game_object)

  return avatar_objects


def create_avatar_objects(num_players,
                          taste_kwargs: Optional[Any] = None,
                          fixed_teams: Optional[bool] = False):
  """Returns list of avatar objects of length 'num_players'."""
  assert num_players % 2 == 0, "num players must be divisible by 2"
  if fixed_teams:
    avatar_objects = _low_vs_high_team_assignment(num_players,
                                                  taste_kwargs=taste_kwargs)
  else:
    avatar_objects = _even_vs_odd_team_assignment(num_players,
                                                  taste_kwargs=taste_kwargs)
  return avatar_objects


def get_config():
  """Default configuration."""
  config = config_dict.ConfigDict()

  # If shaping_kwargs are None then use the default reward structure in which
  # the only positive rewards are those that are delivered when your team
  # captures its opposing team's flag and the only negative rewards are those
  # delivered when the opposing team captures your team's flag. The default
  # reward structure is zero sum.
  config.shaping_kwargs = None

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "READY_TO_SHOOT",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(ACTION_SET))
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(184, 184),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"default"})
  config.default_player_roles = ("default",) * 8

  return config


def build(
    roles: Sequence[str],
    config: config_dict.ConfigDict,
) -> Mapping[str, Any]:
  """Build substrate definition given player roles."""
  num_players = len(roles)
  substrate_definition = dict(
      levelName="paintball__capture_the_flag",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      maxEpisodeLengthFrames=1000,
      spriteSize=8,
      topology="BOUNDED",  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": ASCII_MAP,
          "gameObjects": create_avatar_objects(
              num_players, taste_kwargs=config.shaping_kwargs),
          "scene": create_scene(),
          "prefabs": PREFABS,
          "charPrefabMap": CHAR_PREFAB_MAP,
      },
  )
  return substrate_definition



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests of fixed goal puppeteer."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized

from meltingpot.utils.puppeteers import fixed_goal
from meltingpot.utils.puppeteers import testutils


class FixedGoalTest(parameterized.TestCase):

  def test_goal_sequence(self):
    puppeteer = fixed_goal.FixedGoal(mock.sentinel.goal)
    observations = [{}] * 3
    expected = [mock.sentinel.goal] * 3
    actual, _ = testutils.goals_from_observations(puppeteer, observations)
    self.assertSequenceEqual(actual, expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A simple human player for testing a Melting Pot level.

Use `WASD` keys to move the character around.
Use `Q and E` to turn the character.
Use `SPACE` to fire the zapper.
Use `TAB` to switch between players.
"""

import collections
import enum
from typing import Any, Callable, Dict, Mapping, Optional, Sequence, Tuple

import dm_env
import dmlab2d
from ml_collections import config_dict
import numpy as np
import pygame

from meltingpot.utils.substrates import builder

WHITE = (255, 255, 255)

MOVEMENT_MAP = {
    'NONE': 0,
    'FORWARD': 1,
    'RIGHT': 2,
    'BACKWARD': 3,
    'LEFT': 4,
}


EnvBuilder = Callable[..., dmlab2d.Environment]  # Only supporting kwargs.

ActionMap = Mapping[str, Callable[[], int]]


class RenderType(enum.Enum):
  NONE = 0
  PYGAME = 1


def get_random_direction() -> int:
  """Gets a random direction."""
  return np.random.choice(list(MOVEMENT_MAP.values()))


def get_random_turn() -> int:
  """Gets a random turn."""
  return np.random.choice([-1, 0, 1])


def get_random_fire() -> int:
  """Gets a random fire."""
  return np.random.choice([0, 1])


def get_direction_pressed() -> int:
  """Gets direction pressed."""
  key_pressed = pygame.key.get_pressed()
  if key_pressed[pygame.K_UP] or key_pressed[pygame.K_w]:
    return MOVEMENT_MAP['FORWARD']
  if key_pressed[pygame.K_RIGHT] or key_pressed[pygame.K_d]:
    return MOVEMENT_MAP['RIGHT']
  if key_pressed[pygame.K_DOWN] or key_pressed[pygame.K_s]:
    return MOVEMENT_MAP['BACKWARD']
  if key_pressed[pygame.K_LEFT] or key_pressed[pygame.K_a]:
    return MOVEMENT_MAP['LEFT']
  return MOVEMENT_MAP['NONE']


def get_turn_pressed() -> int:
  """Calculates turn increment."""
  key_pressed = pygame.key.get_pressed()
  if key_pressed[pygame.K_DELETE] or key_pressed[pygame.K_q]:
    return -1
  if key_pressed[pygame.K_PAGEDOWN] or key_pressed[pygame.K_e]:
    return 1
  return 0


def get_space_key_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_SPACE] else 0


def get_key_number_pressed() -> int:
  number_keys = [pygame.K_0, pygame.K_1, pygame.K_2, pygame.K_3, pygame.K_4,
                 pygame.K_5, pygame.K_6, pygame.K_7, pygame.K_8, pygame.K_9]
  for num in range(len(number_keys)):
    if pygame.key.get_pressed()[number_keys[num]]:
      return num
  return -1


def get_key_number_one_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_1] else 0


def get_key_number_two_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_2] else 0


def get_key_number_three_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_3] else 0


def get_key_number_four_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_4] else 0


def get_key_number_five_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_5] else 0


def get_left_control_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_LCTRL] else 0


def get_left_shift_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_LSHIFT] else 0


def get_right_shift_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_RSHIFT] else 0


def get_key_c_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_c] else 0


def get_key_z_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_z] else 0


def get_key_x_pressed() -> int:
  return 1 if pygame.key.get_pressed()[pygame.K_x] else 0


def _split_key(key: str) -> Tuple[str, str]:
  """Splits the key into player index and name."""
  return tuple(key.split('.', maxsplit=1))


def _get_rewards(timestep: dm_env.TimeStep) -> Mapping[str, float]:
  """Gets the list of rewards, one for each player."""
  rewards = {}
  for key in timestep.observation.keys():
    if key.endswith('.REWARD'):
      player_prefix, name = _split_key(key)
      if name == 'REWARD':
        rewards[player_prefix] = timestep.observation[key]
  return rewards


class ActionReader(object):
  """Convert keyboard actions to environment actions."""

  def __init__(self, env: dmlab2d.Environment, action_map: ActionMap):
    # Actions are named "<player_prefix>.<action_name>"
    self._action_map = action_map
    self._action_spec = env.action_spec()
    assert isinstance(self._action_spec, dict)
    self._action_names = set()
    for action_key in self._action_spec.keys():
      _, action_name = _split_key(action_key)
      self._action_names.add(action_name)

  def step(self, player_prefix: str) -> Mapping[str, int]:
    """Update the actions of player `player_prefix`."""
    actions = {action_key: 0 for action_key in self._action_spec.keys()}
    for action_name in self._action_names:
      actions[f'{player_prefix}.{action_name}'] = self._action_map[
          action_name]()
    return actions


def run_episode(
    render_observation: str,
    config_overrides: Dict[str, Any],
    action_map: ActionMap,
    full_config: config_dict.ConfigDict,
    interactive: RenderType = RenderType.PYGAME,
    screen_width: int = 800,
    screen_height: int = 600,
    fps: int = 8,
    verbose_fn: Optional[Callable[[dm_env.TimeStep, int, int], None]] = None,
    text_display_fn: Optional[Callable[[dm_env.TimeStep, int], str]] = None,
    text_font_size: int = 36,
    text_x_pos: int = 20,
    text_y_pos: int = 20,
    text_color: Tuple[int, ...] = WHITE,
    env_builder: EnvBuilder = builder.builder,
    print_events: Optional[bool] = False,
    player_prefixes: Optional[Sequence[str]] = None,
    default_observation: str = 'WORLD.RGB',
    reset_env_when_done: bool = False,
    initial_player_index: int = 0,
) -> None:
  """Run multiplayer environment, with per player rendering and actions.

  This function initialises a Melting Pot environment with the given
  configuration (including possible config overrides), and optionally launches
  the episode as an interactive game using pygame.  The controls are described
  in the action_map, whose keys correspond to discrete actions of the
  environment.

  Args:
    render_observation: A string consisting of the observation name to render.
        Usually 'RGB' for the third person world view.
    config_overrides: A dictionary of settings to override from the original
        `full_config.lab2d_settings`. Typically these are used to set the number
        of players.
    action_map: A dictionary of (discrete) action names to functions that detect
        the keys that correspond to its possible action values.  For example,
        for movement, we might want to have WASD navigation tied to the 'move'
        action name using `get_direction_pressed`.  See examples in the various
        play_*.py scripts.
    full_config: The full configuration for the Melting Pot environment.  These
        usually come from meltingpot/python/configs/environments.
    interactive: A RenderType representing whether the episode should be run
        with PyGame, or without any interface.  Setting interactive to false
        enables running e.g. a random agent via the action_map returning actions
        without polling PyGame (or any human input).  Non interactive runs
        ignore the screen_width, screen_height and fps parameters.
    screen_width: Width, in pixels, of the window to render the game.
    screen_height: Height, in pixels, of the window to render the game.
    fps: Frames per second of the game.
    verbose_fn: An optional function that will be executed for every step of
        the environment.  It receives the environment timestep, a player index
        (will be called for every index), and the current player index. This is
        typically used to print extra information that would be useful for
        debugging a running episode.
    text_display_fn: An optional function for displaying text on screen. It
        receives the environment and the player index, and returns a string to
        display on the pygame screen.
    text_font_size: the font size of onscreen text (from `text_display_fn`)
    text_x_pos: the x position of onscreen text (from `text_display_fn`)
    text_y_pos: the x position of onscreen text (from `text_display_fn`)
    text_color: RGB color of onscreen text (from `text_display_fn`)
    env_builder: The environment builder function to use. By default it is
      meltingpot.builder.
    print_events: An optional bool that if enabled will print events captured
      from the dmlab2d events API on any timestep where they occur.
    player_prefixes: If given, use these as the prefixes of player actions.
      Pressing TAB will cycle through these. If not given, use the standard
      ('1', '2', ..., numPlayers).
    default_observation: Default observation to render if 'render_observation'
      or '{player_prefix}.{render_observation}' is not found in the dict.
    reset_env_when_done: if True, reset the environment once the episode has
      terminated; useful for playing multiple episodes in a row. Note this
      will cause this function to loop infinitely.
    initial_player_index: Initial index of the player to play as. Defaults to 0.
      (Players are always switchable via the tab key.)
  """
  full_config.lab2d_settings.update(config_overrides)
  if player_prefixes is None:
    player_count = full_config.lab2d_settings.get('numPlayers', 1)
    # By default, we use lua indices (which start at 1) as player prefixes.
    player_prefixes = [f'{i+1}' for i in range(player_count)]
  else:
    player_count = len(player_prefixes)
  print(f'Running an episode with {player_count} players: {player_prefixes}.')
  with env_builder(**full_config) as env:

    if len(player_prefixes) != player_count:
      raise ValueError('Player prefixes, when specified, must be of the same '
                       'length as the number of players.')
    player_index = initial_player_index
    timestep = env.reset()

    score = collections.defaultdict(float)
    action_reader = ActionReader(env, action_map)

    if interactive == RenderType.PYGAME:
      pygame.init()
      pygame.display.set_caption('Melting Pot: {}'.format(
          full_config.lab2d_settings.levelName))
      font = pygame.font.SysFont(None, text_font_size)

    scale = 1
    observation_spec = env.observation_spec()
    if render_observation in observation_spec:
      obs_spec = observation_spec[render_observation]
    elif f'1.{render_observation}' in observation_spec:
      # This assumes all players have the same observation, which is true for
      # MeltingPot environments.
      obs_spec = observation_spec[f'1.{render_observation}']
    else:
      # Falls back to 'default_observation.'
      obs_spec = observation_spec[default_observation]

    observation_shape = obs_spec.shape
    observation_height = observation_shape[0]
    observation_width = observation_shape[1]
    scale = min(screen_height // observation_height,
                screen_width // observation_width)
    if interactive == RenderType.PYGAME:
      game_display = pygame.display.set_mode(
          (observation_width * scale, observation_height * scale))
      clock = pygame.time.Clock()
    stop = False

    # Game loop
    while True:

      # Check for pygame controls
      if interactive == RenderType.PYGAME:
        for event in pygame.event.get():
          if event.type == pygame.QUIT:
            stop = True

          if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_TAB:
              player_index = (player_index + 1) % player_count
            break
      player_prefix = player_prefixes[player_index] if player_prefixes else ''

      if stop:
        break

      # Compute next timestep
      actions = action_reader.step(player_prefix) if player_count else []
      timestep = env.step(actions)
      if timestep.step_type == dm_env.StepType.LAST:
        if reset_env_when_done:
          timestep = env.reset()
        else:
          break

      rewards = _get_rewards(timestep)
      for i, prefix in enumerate(player_prefixes):
        if verbose_fn:
          verbose_fn(timestep, i, player_index)
        score[prefix] += rewards[prefix]
        if i == player_index and rewards[prefix] != 0:
          print(f'Player {prefix} Score: {score[prefix]}')

      # Print events if applicable
      if print_events and hasattr(env, 'events'):
        events = env.events()
        # Only print events on timesteps when there are events to print.
        if events:
          print(events)

      # pygame display
      if interactive == RenderType.PYGAME:
        # show visual observation
        if render_observation in timestep.observation:
          obs = timestep.observation[render_observation]
        elif f'{player_prefix}.{render_observation}' in timestep.observation:
          obs = timestep.observation[f'{player_prefix}.{render_observation}']
        else:
          # Fall back to default_observation.
          obs = timestep.observation[default_observation]
        obs = np.transpose(obs, (1, 0, 2))  # PyGame is column major!

        surface = pygame.surfarray.make_surface(obs)
        rect = surface.get_rect()

        surf = pygame.transform.scale(
            surface, (rect[2] * scale, rect[3] * scale))
        game_display.blit(surf, dest=(0, 0))

        # show text
        if text_display_fn:
          if player_count == 1:
            text_str = text_display_fn(timestep, 0)
          else:
            text_str = text_display_fn(timestep, player_index)
          img = font.render(text_str, True, text_color)
          game_display.blit(img, (text_x_pos, text_y_pos))

        # tick
        pygame.display.update()
        clock.tick(fps)

    if interactive == RenderType.PYGAME:
      pygame.quit()
    for prefix in player_prefixes:
      print('Player %s: score is %g' % (prefix, score[prefix]))



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for game_object_utils."""

from absl.testing import absltest
from absl.testing import parameterized

from meltingpot.utils.substrates import game_object_utils


def get_transform(x, y, orientation):
  return game_object_utils.Transform(position=game_object_utils.Position(x, y),
                                     orientation=orientation)


class ParseMapTest(parameterized.TestCase):

  @parameterized.parameters(
      ('\nHello', 'H', 1),
      ('\nHello', 'h', 0),
      ('\nHello', 'l', 2),
      ('\nHello\nWorld', 'l', 3),
      ('\nHello\nWorld', 'o', 2),
      ('\nHello\nWorld', 'd', 1),
      ('\nHello\nWorld', 'W', 1),
      ('\nWWWW\nW AW\nWWWW', 'A', 1),
      ('\nWWWW\nW AW\nWWWW', 'W', 10),
      ('\nWWWW\nW AW\nWWWW', 'P', 0),
      )
  def test_get_positions_length(self, ascii_map, char, exp_len):
    transforms = game_object_utils.get_game_object_positions_from_map(
        ascii_map, char)
    self.assertLen(transforms, exp_len)

  def test_get_positions(self):
    # Locations of 'A' -> (2, 1)
    # Locations of ' ' -> (1, 1), (3, 1) and (4, 1)
    ascii_map = '''
WWWWWW
W A  W
WWWWWW
'''
    transforms = game_object_utils.get_game_object_positions_from_map(
        ascii_map, 'A')
    self.assertSameElements(
        [get_transform(2, 1, game_object_utils.Orientation.NORTH)], transforms)

    transforms = game_object_utils.get_game_object_positions_from_map(
        ascii_map, ' ')
    self.assertSameElements(
        [
            get_transform(1, 1, game_object_utils.Orientation.NORTH),
            get_transform(3, 1, game_object_utils.Orientation.NORTH),
            get_transform(4, 1, game_object_utils.Orientation.NORTH)
        ],
        transforms)

    transforms = game_object_utils.get_game_object_positions_from_map(
        ascii_map, 'W')
    self.assertSameElements(
        [
            # Top walls
            get_transform(0, 0, game_object_utils.Orientation.NORTH),
            get_transform(1, 0, game_object_utils.Orientation.NORTH),
            get_transform(2, 0, game_object_utils.Orientation.NORTH),
            get_transform(3, 0, game_object_utils.Orientation.NORTH),
            get_transform(4, 0, game_object_utils.Orientation.NORTH),
            get_transform(5, 0, game_object_utils.Orientation.NORTH),

            # Side walls
            get_transform(0, 1, game_object_utils.Orientation.NORTH),
            get_transform(5, 1, game_object_utils.Orientation.NORTH),

            # Bottom walls
            get_transform(0, 2, game_object_utils.Orientation.NORTH),
            get_transform(1, 2, game_object_utils.Orientation.NORTH),
            get_transform(2, 2, game_object_utils.Orientation.NORTH),
            get_transform(3, 2, game_object_utils.Orientation.NORTH),
            get_transform(4, 2, game_object_utils.Orientation.NORTH),
            get_transform(5, 2, game_object_utils.Orientation.NORTH),
        ],
        transforms)

  def test_get_game_objects(self):
    ascii_map = '''
WWWWWW
W A  W
WWWWWW
'''
    wall = {
        'name': 'wall',
        'components': [
            {
                'component': 'PieceTypeManager',
                'kwargs': {
                    'initialPieceType': 'wall',
                    'pieceTypeConfigs': [{'pieceType': 'wall',}],
                },
            },
            {
                'component': 'Transform',
                'kwargs': {
                    'position': (0, 0),
                    'orientation': 'N'
                },
            },
        ]
    }
    apple = {
        'name': 'apple',
        'components': [
            {
                'component': 'PieceTypeManager',
                'kwargs': {
                    'initialPieceType': 'apple',
                    'pieceTypeConfigs': [{'pieceType': 'apple',}],
                },
            },
            {
                'component': 'Transform',
                'kwargs': {
                    'position': (0, 0),
                    'orientation': 'N'
                },
            },
        ]
    }
    prefabs = {'wall': wall, 'apple': apple}
    game_objects = game_object_utils.get_game_objects_from_map(
        ascii_map, {'W': 'wall', 'A': 'apple'}, prefabs)
    self.assertLen(game_objects, 15)
    self.assertEqual(
        1,
        sum([1 if go['name'] == 'apple' else 0 for go in game_objects]))
    self.assertEqual(
        14,
        sum([1 if go['name'] == 'wall' else 0 for go in game_objects]))

    positions = []
    for go in game_objects:
      if go['name'] == 'wall':
        positions.append(game_object_utils.get_first_named_component(
            go, 'Transform')['kwargs']['position'])

    self.assertSameElements(
        [
            (0, 0), (1, 0), (2, 0), (3, 0), (4, 0), (5, 0),  # Top walls
            (0, 1), (5, 1),  # Side walls
            (0, 2), (1, 2), (2, 2), (3, 2), (4, 2), (5, 2),  # Bottom walls
        ],
        positions)


AVATAR = {
    'name': 'avatar',
    'components': [
        {
            'component': 'StateManager',
            'kwargs': {
                'initialState': 'player',
                'stateConfigs': [
                    {'state': 'player',
                     'layer': 'upperPhysical',
                     'sprite': 'Avatar',},  # Will be overridden

                    {'state': 'playerWait',},
                ]
            }
        },
        {
            'component': 'Transform',
            'kwargs': {
                'position': (0, 0),
                'orientation': 'N'
            }
        },
        {
            'component': 'Appearance',
            'kwargs': {
                'renderMode': 'ascii_shape',
                'spriteNames': ['Avatar'],  # Will be overridden
                'spriteShapes': ["""*"""],
                'palettes': [(0, 0, 255, 255)],  # Will be overridden
                'noRotates': [True]
            }
        },
        {
            'component': 'Avatar',
            'kwargs': {
                'index': -1,  # Will be overridden
                'spawnGroup': 'spawnPoints',
                'aliveState': 'player',
                'waitState': 'playerWait',
                'actionOrder': ['move'],
                'actionSpec': {
                    'move': {'default': 0, 'min': 0, 'max': 4},
                },
            }
        },
    ],
}


BADGE = {
    'name': 'avatar_badge',
    'components': [
        {
            'component': 'StateManager',
            'kwargs': {
                'initialState': 'badgeWait',
                'stateConfigs': [
                    {'state': 'badge',
                     'layer': 'overlay',
                     'sprite': 'Badge',
                     'groups': ['badges']},

                    {'state': 'badgeWait',
                     'groups': ['badgeWaits']},
                ]
            }
        },
        {
            'component': 'Transform',
            'kwargs': {
                'position': (0, 0),
                'orientation': 'N'
            }
        },
        {
            'component': 'Appearance',
            'kwargs': {
                'renderMode': 'ascii_shape',
                'spriteNames': ['Badge'],
                'spriteShapes': ['*'],
                'palettes': [(0, 0, 255, 255)],
                'noRotates': [False]
            }
        },
        {
            'component': 'AvatarConnector',
            'kwargs': {
                'playerIndex': -1,  # player index to be overwritten.
                'aliveState': 'badge',
                'waitState': 'badgeWait'
            }
        },
    ]
}


class BuildAvatarObjectsTest(parameterized.TestCase):

  @parameterized.parameters(
      [1], [2], [3], [4], [5]
      )
  def test_simple_build(self, num_players):
    prefabs = {'avatar': AVATAR}
    avatars = game_object_utils.build_avatar_objects(
        num_players=num_players,
        prefabs=prefabs,
        player_palettes=None,
        )
    self.assertLen(avatars, num_players)

  def test_with_palette_build(self):
    palettes = [(255, 0, 0, 255), (0, 255, 0, 255)]
    prefabs = {'avatar': AVATAR}
    avatars = game_object_utils.build_avatar_objects(
        num_players=2,
        prefabs=prefabs,
        player_palettes=palettes,
        )
    self.assertLen(avatars, 2)
    self.assertEqual(
        game_object_utils.get_first_named_component(
            avatars[0], 'Appearance')['kwargs']['palettes'][0],
        palettes[0])
    self.assertEqual(
        game_object_utils.get_first_named_component(
            avatars[1], 'Appearance')['kwargs']['palettes'][0],
        palettes[1])


class BuildAvatarBadgesTest(parameterized.TestCase):

  @parameterized.parameters(
      [1], [2], [3], [4], [5]
      )
  def test_simple_build(self, num_players):
    prefabs = {'avatar_badge': BADGE}
    badges = game_object_utils.build_avatar_badges(
        num_players=num_players,
        prefabs=prefabs,
        badge_palettes=None,
        )
    self.assertLen(badges, num_players)

  def test_with_palette_build(self):
    palettes = [(255, 0, 0, 255), (0, 255, 0, 255)]
    prefabs = {'avatar_badge': BADGE}
    badges = game_object_utils.build_avatar_badges(
        num_players=2,
        prefabs=prefabs,
        badge_palettes=palettes,
        )
    self.assertLen(badges, 2)
    self.assertEqual(
        game_object_utils.get_first_named_component(
            badges[0], 'Appearance')['kwargs']['palettes'][0],
        palettes[0])
    self.assertEqual(
        game_object_utils.get_first_named_component(
            badges[1], 'Appearance')['kwargs']['palettes'][0],
        palettes[1])


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for gift_refinements puppeteers."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized

from meltingpot.utils.puppeteers import gift_refinements
from meltingpot.utils.puppeteers import testutils

_COLLECT = mock.sentinel.collect
_CONSUME = mock.sentinel.consume
_GIFT = mock.sentinel.gift


class GiftRefinementsCooperatorTest(parameterized.TestCase):

  @parameterized.parameters(
      [(0, 0, 0), _COLLECT],
      [(0, 0, 1), _CONSUME],
      [(1, 0, 0), _GIFT],
      [(2, 0, 0), _GIFT],
      [(2, 0, 2), _CONSUME],
      [(1, 1, 0), _CONSUME],
      [(1, 1, 1), _CONSUME],
      [(5, 2, 0), _CONSUME],
      [(5, 5, 5), _CONSUME],
  )
  def test(self, inventory, expected):
    puppeteer = gift_refinements.GiftRefinementsCooperator(
        collect_goal=_COLLECT,
        consume_goal=_CONSUME,
        gift_goal=_GIFT,
    )
    (actual,), _ = testutils.goals_from_observations(
        puppeteer, [{'INVENTORY': inventory}])
    self.assertEqual(actual, expected)


class GiftRefinementsExtremeCooperatorTest(parameterized.TestCase):

  @parameterized.parameters(
      [(0, 0, 0), _COLLECT],
      [(0, 0, 1), _CONSUME],
      [(1, 0, 0), _GIFT],
      [(2, 0, 2), _CONSUME],
      [(1, 1, 0), _GIFT],
      [(1, 1, 1), _CONSUME],
      [(5, 2, 0), _GIFT],
      [(5, 5, 5), _CONSUME],
  )
  def test(self, inventory, expected):
    puppeteer = gift_refinements.GiftRefinementsExtremeCooperator(
        collect_goal=_COLLECT,
        consume_goal=_CONSUME,
        gift_goal=_GIFT,
    )
    (actual,), _ = testutils.goals_from_observations(
        puppeteer, [{'INVENTORY': inventory}])
    self.assertEqual(actual, expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl.testing import absltest
import dm_env
import numpy as np

from meltingpot.testing import mocks
from meltingpot.testing import substrates
from meltingpot.utils.substrates import specs as meltingpot_specs
from meltingpot.utils.substrates import substrate


class MocksTest(substrates.SubstrateTestCase):

  def test_value_from_specs(self):
    specs = (
        {'a': dm_env.specs.Array([1, 2, 3], dtype=np.uint8)},
        {'b': dm_env.specs.Array([1, 2, 3], dtype=np.uint8)},
    )
    actual = mocks._values_from_specs(specs)
    expected = (
        {'a': np.zeros([1, 2, 3], dtype=np.uint8) + 0},
        {'b': np.zeros([1, 2, 3], dtype=np.uint8) + 1},
    )
    np.testing.assert_equal(actual, expected)

  def test_mock_substrate(self):
    num_players = 2
    num_actions = 3
    observation_spec = {'a': dm_env.specs.Array([], dtype=np.uint8)}
    mock = mocks.build_mock_substrate(
        num_players=num_players,
        num_actions=num_actions,
        observation_spec=observation_spec)

    expected_observation = ({'a': np.uint8()}, {'a': np.uint8() + 1})
    expected_reward = tuple(float(n) for n in range(num_players))

    with self.subTest('is_substrate'):
      self.assertIsInstance(mock, substrate.Substrate)
    with self.subTest('error_getting_invalid'):
      with self.assertRaises(AttributeError):
        mock.no_such_method()  # pytype: disable=attribute-error
    with self.subTest('error_setting_invalid'):
      with self.assertRaises(AttributeError):
        mock.no_such_method = None

    with self.subTest('can_enter_context'):
      with mock as c:
        self.assertEqual(c.discount_spec(), mock.discount_spec())

    with self.subTest('action_spec'):
      self.assertEqual(mock.action_spec(), (meltingpot_specs.action(3),) * 2)
    with self.subTest('reward_spec'):
      self.assertLen(mock.reward_spec(), num_players)
    with self.subTest('observation_spec'):
      self.assertLen(mock.observation_spec(), num_players)

    with self.subTest('reset'):
      expected = dm_env.TimeStep(
          step_type=dm_env.StepType.FIRST,
          observation=expected_observation,
          reward=(0.,) * num_players,
          discount=0.,
      )
      self.assertEqual(mock.reset(), expected)
    with self.subTest('step'):
      expected = dm_env.transition(expected_reward, expected_observation)
      self.assertEqual(mock.step([0, 0]), expected)
    with self.subTest('events'):
      self.assertEmpty(mock.events())
    with self.subTest('observation'):
      self.assertEqual(mock.observation(), expected_observation)

  def test_mock_substrate_like(self):
    mock = mocks.build_mock_substrate_like('clean_up')
    self.assert_step_matches_specs(mock)

  def test_mock_scenario_like(self):
    mock = mocks.build_mock_scenario_like('clean_up_0')
    self.assert_step_matches_specs(mock)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Territory: Inside Out.

Example video: https://youtu.be/LdbIjnHaisU

See _Territory: Open_ for the general description of the mechanics at play in
this substrate.

In this substrate, _Territory: Inside Out_,  players start on the outside of
a randomly generated maze of resources. They must move from their starting
locations inward toward the center of the map to claim territory. In so doing
they will quickly encounter their coplayers who will be doing the same thing
from their own starting locations. In order to get high scores, agents must be
able to rapidly negotiate tacit agreements with one another concerning the
borders between their respective territories. Since the spatial arrangement of
the resources differs from episode to episode, so too does the negotiation
problem to be solved.
"""

from ml_collections import config_dict

from meltingpot.configs.substrates import territory as base_config
from meltingpot.utils.substrates import map_helpers
from meltingpot.utils.substrates import specs

build = base_config.build

ASCII_MAP = """
F=====================T
|,,,,,,,,,,P,,,,,,,,,,|
|,P,,,,QQ,,,,,QQ,,,,P,|
|,,RRR,,,,RRR,,,,RRR,,|
|,,R,RAAAAR,RAAAAR,R,,|
|,,RRR,BB,RRR,BB,RRR,,|
|,,,A,,BB,,A,,BB,,A,,,|
|,Q,ABBRRBBABBRRBBA,Q,|
|,Q,ABBRRBBABBRRBBA,Q,|
|,,,A,,BB,,A,,BB,,A,,,|
|,,RRR,BB,RRR,BB,RRR,,|
|P,R,RAAAAR,RAAAAR,R,P|
|,,RRR,BB,RRR,BB,RRR,,|
|,,,A,,BB,,A,,BB,,A,,,|
|,Q,ABBRRBBABBRRBBA,Q,|
|,Q,ABBRRBBABBRRBBA,Q,|
|,,,A,,BB,,A,,BB,,A,,,|
|,,RRR,BB,RRR,BB,RRR,,|
|,,R,RAAAAR,RAAAAR,R,,|
|,,RRR,,,,RRR,,,,RRR,,|
|,P,,,,QQ,,,,,QQ,,,,P,|
|,,,,,,,,,,P,,,,,,,,,,|
L=====================J
"""

# `prefab` determines which prefab game object to use for each `char` in the
# ascii map.
resource_associated_prefabs = ["floor", "resource_texture", "resource",
                               "reward_indicator", "damage_indicator"]
resource = {"type": "all", "list": resource_associated_prefabs}
spawn_point_associated_prefabs = ["floor", "spawn_point"]
spawn_point = {"type": "all", "list": spawn_point_associated_prefabs}
CHAR_PREFAB_MAP = {
    "P": spawn_point,
    "Q": map_helpers.a_or_b_with_odds(spawn_point, "floor", odds=(1, 6)),
    ",": "floor",
    "F": {"type": "all", "list": ["wall", "wall_highlight_nw"]},
    "|": {"type": "all", "list": ["wall", "wall_highlight_e_w"]},
    "=": {"type": "all", "list": ["wall", "wall_highlight_n_s"]},
    "T": {"type": "all", "list": ["wall", "wall_highlight_ne"]},
    "J": {"type": "all", "list": ["wall", "wall_highlight_se"]},
    "L": {"type": "all", "list": ["wall", "wall_highlight_sw"]},
    "R": resource,
    "A": map_helpers.a_or_b_with_odds(resource, "floor", odds=(2, 1)),
    "B": map_helpers.a_or_b_with_odds(resource, "floor", odds=(1, 3)),
}


def get_config():
  """Default configuration."""
  config = base_config.get_config()

  # Override the map layout settings.
  config.layout = config_dict.ConfigDict()
  config.layout.ascii_map = ASCII_MAP
  config.layout.char_prefab_map = CHAR_PREFAB_MAP
  config.layout.topology = "BOUNDED"

  # The specs of the environment (from a single-agent perspective).
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(184, 184),
  })

  # The roles assigned to each player.
  config.default_player_roles = ("default",) * 5

  return config
